/******************************************************************************/
/*                                                                            */
/*                                cluster.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                              Aug 2003 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/08/25 13:31:50 $
 *   $Revision: 1.4 $
 *
 */

#define _cluster_c_

#include <cluster.h>

/* ------------------------------------------------- */
/* ----- int cluster_init(unsigned short, int) ----- */
/* ------------------------------------------------- */
/*
 * Initialize workspace for clustering.
 */
int cluster_init(unsigned short dim, int fullc)
{

#ifdef HAVE_GSL
  if (fullc) {
    gsl_set_error_handler_off();
    
    if ((_S = gsl_matrix_alloc(dim, dim)) == NULL)
      return(1);
    
    if ((_ws = gsl_eigen_symm_alloc(dim)) == NULL) {
      cluster_reset();
      return(1);
    }
    
    if ((_ev = gsl_vector_calloc(dim)) == NULL) {
      cluster_reset();
      return(1);
    }
    _ctype = CLUSTER_FULL_COV;
  }
#else
  if (fullc) {
    fprintf(stderr, "cluster_init() error -- cannot initialize clustering with full covariance\n");
    fprintf(stderr, "                        build audioseg with GSL support for full covariance\n");
    return(1);
  }
#endif

  return(0);
}

/* ------------------------------------ */
/* ----- void cluster_reset(void) ----- */
/* ------------------------------------ */
/*
 * Reset clustering workspace.
 */
void cluster_reset(void)
{
#ifdef HAVE_GSL
  if (_ctype == CLUSTER_FULL_COV) {
    if (_S)
      gsl_matrix_free(_S);
    if (_ev)
      gsl_vector_free(_ev);
    if (_ws)
      gsl_eigen_symm_free(_ws);
  }
#endif
}

/* ---------------------------------------------------- */
/* ----- cluster_t *cluster_alloc(unsigned short) ----- */
/* ---------------------------------------------------- */
/*
 * Allocate memory for a cluster with a given dimension.
 */
cluster_t *cluster_alloc(unsigned short dim)
{
  unsigned short j;
  cluster_t *p;

  if ((p = (cluster_t *)malloc(sizeof(cluster_t))) == NULL)
    return(NULL);

  p->dim = dim;
  p->nframes = 0;
  p->m = NULL;
  p->v = NULL;
  p->nsegs = 0;
  p->link = NULL;
  p->next = NULL;
  p->prev = NULL;

  if (dim) {
    if ((p->m = (double *)malloc(dim * sizeof(double))) == NULL) {
      cluster_free(p);
      return(NULL);
    }
    
    if (_ctype == CLUSTER_FULL_COV) {

      if ((p->v = (double **)malloc(dim * sizeof(double *))) == NULL) {
	cluster_free(p);
	return(NULL);
      }
      for (j = 0; j < dim; j++)
	*(p->v+j) = NULL;

      for (j = 0; j < dim; j++) {
	if ((*(p->v+j) = (double *)malloc((dim - j) * sizeof(double))) == NULL) {
	  cluster_free(p);
	  return(NULL);
	}
      }
    }
    else { /* CLUSTER_DIAG_COV */
      if ((p->v = (double **)malloc(sizeof(double *))) == NULL) {
	cluster_free(p);
	return(NULL);
      }
      if ((*(p->v) = (double *)malloc(dim * sizeof(double))) == NULL) {
	  cluster_free(p);
	  return(NULL);
      }
    }
  }

  return(p);
}

/* ------------------------------------------ */
/* ----- void cluster_free(cluster_t *) ----- */
/* ------------------------------------------ */
/*
 * Free memory allocated for a cluster.
 */
void cluster_free(cluster_t *p)
{
  unsigned short j;
  seglink_t *link, *tmp;

  if (p) {

    if (p->m)
      free(p->m);

    if (p->v) {

      if (_ctype == CLUSTER_FULL_COV) {
	for (j = 0; j < p->dim; j++)
	  if (*(p->v+j))
	    free(*(p->v+j));
      }
      else if (*(p->v))
	free(*(p->v));

      free(p->v);
    }

    if (p->link) {
      link = p->link;

      while (link) {
	tmp = link->next;
	free(link);
	link = tmp;
      }
    }
  }
}

/* ----------------------------------------------- */
/* ----- void cluster_list_free(cluster_t *) ----- */
/* ----------------------------------------------- */
/*
 * Free memory allocated for a cluster list.
 */
void cluster_list_free(cluster_t *listp)
{
  cluster_t *next, *p = listp;

  while (p) {
    next = p->next;
    cluster_free(p);
    p = next;
  }
}

/* --------------------------------------------------------------------------- */
/* ----- cluster_t *cluster_merge(cluster_t *, cluster_t *, cluster_t *) ----- */
/* --------------------------------------------------------------------------- */
/*
 * Merge the two clusters into the first one.  Return a pointer to
 * the, possibly changed, list address.  
 */
cluster_t *cluster_merge(cluster_t *listp, cluster_t *c1, cluster_t *c2)
{
  unsigned short i, j, dim = c1->dim;
  double *mp1, *vp1;
  double *mp2, *vp2;
  cluster_t *p = listp;
  seglink_t *linkp;

  /* merge accumulators */
  mp1 = c1->m;
  mp2 = c2->m;

  c1->nframes += c2->nframes;

  if (_ctype == CLUSTER_FULL_COV) {

    for (i = 0; i < dim; i++) {
      *(mp1+i) += *(mp2+i);

      vp1 = *(c1->v+i);
      vp2 = *(c2->v+i);

      for (j = 0; j < dim - i; j++)
	*(vp1+j) += *(vp2+j);
    }
  }
  else {
    vp1 = *(c1->v);
    vp2 = *(c2->v);

    for (i = 0; i < dim; i++) {
      *(mp1+i) += *(mp2+i);
      *(vp1+i) += *(vp2+i);
    }
  }

  /* compute new determinant */
  cluster_const_set(c1);

  /* merge links */
  linkp = c1->link;
  while (linkp->next)
    linkp = linkp->next;

  linkp->next = c2->link;
  c2->link = NULL; /* to prevent memory release of c2 link in cluster_free() */

  c1->nsegs += c2->nsegs;

  /* remove second cluster from list */
  if (c2->prev)
    c2->prev->next = c2->next;
  else
    p = c2->next;

  if (c2->next)
    c2->next->prev = c2->prev;

  cluster_free(c2);

  return(p);
}

/* ----------------------------------------------- */
/* ----- void cluster_const_set(cluster_t *) ----- */
/* ----------------------------------------------- */
/*
 * Set cluster's log determinant.
 *
 * NOTE: only the lower triangular part of _S is used in
 * gsl_eigen_symm() so there's no need to fill in the upper part. 
 */
double cluster_const_set(cluster_t *c)
{
  double z, zz, s;
  unsigned short i, dim = c->dim;
  double *m, *v;
#if HAVE_GSL
  unsigned short j;
  double *p;
#endif

  m = c->m;
  z = (double)1.0 / (double)(c->nframes);
  zz = z * z;

#if HAVE_GSL  
  if (_ctype == CLUSTER_FULL_COV) {

    for (i = 0; i < dim; i++) {
      p = gsl_matrix_ptr(_S, i, i);
      v = *(c->v+i);
      
      *p = *v * z - *(m+i) * *(m+i) * zz;
      for (j = 1; j < dim - i; j++) {
	p += dim;
	*p = *(v+j) * z - (*(m+i) * *(m+i+j)) * zz;
      }
    }

    gsl_eigen_symm(_S, _ev, _ws);
    
    p = gsl_vector_ptr(_ev, 0);
    s = 0.0;
    for (i = 0; i < dim; i++)
      s += log(*(p+i));
  }
  else {
    v = *(c->v);
    s = 0.0;

    for (i = 0; i < dim; i++)
      s += log(*(v+i) * z - *(m+i) * *(m+i) * zz);
  }
#else
  v = *(c->v);
  s = 0.0;

  for (i = 0; i < dim; i++)
    s += log(*(v+i) * z - *(m+i) * *(m+i) * zz);
#endif

  c->logdet = s;

  return(s);
}

/* ------------------------------------------------------------- */
/* ----- icdistmat_t *distance_matrix_alloc(unsigned long) ----- */
/* ------------------------------------------------------------- */
/*
 * Allocate memory for the inter-cluster distance matrix.
 */
icdistmat_t *distance_matrix_alloc(unsigned long nmax)
{
  icdistmat_t *p;
  unsigned long i, j;
  struct _icdistrow_s *rowp;

  if ((p = (icdistmat_t *)malloc(sizeof(icdistmat_t))) == NULL)
    return(NULL);

  p->n = 0;
  p->nmax = 0;
  p->row = NULL;

  if ((p->row = (struct _icdistrow_s *)malloc((nmax - 1) * sizeof(struct _icdistrow_s))) == NULL) {
    distance_matrix_free(p);
    return(NULL);
  }  
  for (i = 0; i < nmax - 1; i++) {
    rowp = p->row + i;
    rowp->c1 = NULL;
    rowp->n = 0;
    rowp->elt = NULL;
  }
  p->nmax = nmax;

  for (i = 0; i < nmax - 1; i++) {
    rowp = p->row + i;

    if ((rowp->elt = (icdist_t *)malloc((nmax - i - 1) * sizeof(icdist_t))) == NULL) {
      distance_matrix_free(p);
      return(NULL);
    }

    for (j = 0; j < nmax - i - 1; j++) {
      rowp->elt[j].c2 = NULL;
      rowp->elt[j].flag = 0;
      rowp->elt[j].d = 0;
    }
  }
  
  return(p);
}

/* ---------------------------------------------------- */
/* ----- void distance_matrix_free(icdistmat_t *) ----- */
/* ---------------------------------------------------- */
/*
 * Free memory allocated to a distance matrix.
 */
void distance_matrix_free(icdistmat_t *p)
{
  unsigned long i;

  if (p) {
    if (p->row) {
      for (i = 0; i < p->nmax - 1; i++) 
	if (p->row[i].elt)
	  free(p->row[i].elt);
      free(p->row);
    }
    free(p);
  }
}

/* ------------------------------------------------------------------------------------------- */
/* ----- void distance_matrix_fill(icdistmat_t *, cluster_t *[2], double *, cluster_t *) ----- */
/* -----                           double(*)(cluster_t *, cluster_t *, float), float)    ----- */
/* ------------------------------------------------------------------------------------------- */
void distance_matrix_fill(icdistmat_t *mat, cluster_t *c[2], double *dmin, cluster_t *listp,
			  double(*dist)(cluster_t *, cluster_t *, float), float lambda)
{
  unsigned long i, j, ibest, jbest;
  cluster_t *p, *pp;
  double icd;
  icdist_t *row;

  i = 0;
  p = listp;
  *dmin = DBL_MAX;
  c[0] = c[1] = NULL;
  ibest = jbest = 0; /* makes gcc happy! */

  while (p) {
    
    j = 0;
    pp = p->next;

    if (! pp)
      break;

    row = mat->row[i].elt;

    while (pp) {
      icd = dist(p, pp, lambda);
      
      if (icd < *dmin) {
	*dmin = icd;
	c[0] = p;
	c[1] = pp;
	ibest = i;
	jbest = j;
      }

      (row+j)->c2 = pp;
      (row+j)->flag = 0;
      (row+j)->d = icd;

      j += 1;
      pp = pp->next;
    }

    mat->row[i].c1 = p;
    mat->row[i].n = j;

    i += 1;
    p = p->next;  
  }

  mat->n = i;
  
  mat->row[ibest].elt[jbest].flag = 1; /* flag elt (ibest, jbest) */
}

/* ------------------------------------------------------------------------------- */
/* ----- int distance_matrix_search(icdistmat_t *, cluster_t *[2], double *) ----- */
/* ------------------------------------------------------------------------------- */
/* 
 * Search in inter-cluster matrix the univisited pair with the
 * smallest distance.
 */
int distance_matrix_search(icdistmat_t *mat, cluster_t *c[2], double *dmin)
{
  int found = 0;
  unsigned long i, j, ibest, jbest, ncols;
  icdist_t *p;

  *dmin = DBL_MAX;
  c[0] = c[1] = NULL;
  ibest = jbest = 0; /* makes gcc happy! */
  
  for (i = 0; i < mat->n; i++) {
    ncols = mat->row[i].n;
    p = mat->row[i].elt;
    
    for (j = 0; j < ncols; j++) {

      if ((p+j)->flag)
	continue;

      if ((p+j)->d < *dmin) {
	*dmin = (p+j)->d;
	ibest = i;
	jbest = j;
	found = 1;
      }
    }
  }

  c[0] = mat->row[ibest].c1;
  c[1] = mat->row[ibest].elt[jbest].c2;
  mat->row[ibest].elt[jbest].flag = 1; /* flag elt (ibest, jbest) */
  
  return(found);
}

/* ------------------------------------------------------------------------ */
/* ----- double cluster_kl2_distance(cluster_t *, cluster_t *, float) ----- */
/* ------------------------------------------------------------------------ */
/*
 * Compute symmetric KL distance between two clusters.
 */
double cluster_kl2_distance(cluster_t *c1, cluster_t *c2, float lambda)
{
  double *m1, *m2, **v1, **v2;
  unsigned short i, dim = c1->dim;
  double d = 0.0;
  double a, mi1, mi2, vi1, vi2, iv1, iv2;
  double z1, z2;

  m1 = c1->m; v1 = c1->v;
  m2 = c2->m; v2 = c2->v;
  
  z1 = (double)1.0 / (double)(c1->nframes);
  z2 = (double)1.0 / (double)(c2->nframes);

  for (i = 0; i < dim; i++) {
    mi1 = *(m1+i) * z1;
    mi2 = *(m2+i) * z2;
    a = mi1 - mi2;
    
    if (_ctype == CLUSTER_FULL_COV) {
      /* covariance matrices are upper diagonal - 1st element of a row
	 is the diagonal element */
      vi1 = **(v1+i) * z1 - mi1 * mi1; 
      vi2 = **(v2+i) * z2 - mi2 * mi2;
    }
    else { /* CLUSTER_DIAG_COV */
      /* c1->v[0] and c2->v[0] are covariance vectors */
      vi1 = *((*v1)+i) * z1 - mi1 * mi1; 
      vi2 = *((*v2)+i) * z2 - mi2 * mi2;
    }

    iv1 = (double)1.0 / vi1;
    iv2 = (double)1.0 / vi2;
    
    d = d + (vi1 * iv2 + vi2 * iv1) + a * a * (iv1 + iv2) - 2;
  }

  return(d / (double)dim);
}

/* ------------------------------------------------------------------------- */
/* ----- double cluster_like_distance(cluster_t *, cluster_t *, float) ----- */
/* ------------------------------------------------------------------------- */
/*
 * Compute likelihood ratio distance netween two clusters.
 *
 * NOTE: only the lower triangular part of _S is used in
 * gsl_eigen_symm() so there's no need to fill in the upper part. 
 */
double cluster_like_distance(cluster_t *c1, cluster_t *c2, float lambda)
{
  double a, z, zz, d, s;
  unsigned long n;
  unsigned short i, dim = c1->dim;
  double *m1, *m2, *v2, *v1;
#if HAVE_GSL
  unsigned short j;
  double *p;
#endif

  m1 = c1->m;
  m2 = c2->m;
  
  /* compute ldet12 */
  n = c1->nframes + c2->nframes;
  z = (double)1.0 / (double)n;
  zz = z * z;
  
#if HAVE_GSL
  if (_ctype == CLUSTER_FULL_COV) {

    for (i = 0; i < dim; i++) {
      p = gsl_matrix_ptr(_S, i, i);
      v1 = *(c1->v+i);
      v2 = *(c2->v+i);

      a = *(m1+i) + *(m2+i);

      *p = (*v1 + *v2) * z - a * a * zz;
      
      for (j = 1; j < dim - i; j++) {
	p += dim;
	*p = (*(v1+j) + *(v2+j)) * z - a * (*(m1+i+j) + *(m2+i+j)) * zz;
      }
    }

    gsl_eigen_symm(_S, _ev, _ws);

    p = gsl_vector_ptr(_ev, 0);
    s = 0.0;
    for (i = 0; i < dim; i++)
      s += log(*(p+i));
  }
  else { /* CLUSTER_DIAG_COV */
    v1 = *(c1->v);
    v2 = *(c2->v);
    s = 0.0;

    for (i = 0; i < dim; i++) {
      a = *(m1+i) + *(m2+i);
      s += log((*(v1+i) + *(v2+i)) * z - a * a * zz);
    }
  }
#else
  v1 = *(c1->v);
  v2 = *(c2->v);
  s = 0.0;

  for (i = 0; i < dim; i++) {
    a = *(m1+i) + *(m2+i);
    s += log((*(v1+i) + *(v2+i)) * z - a * a * zz);
  }
#endif

  d = n * s - c1->nframes * c1->logdet - c2->nframes * c2->logdet;
    
  return(0.5 * d);
}

/* ------------------------------------------------------------------------ */
/* ----- double cluster_bic_distance(cluster_t *, cluster_t *, float) ----- */
/* ------------------------------------------------------------------------ */
/*
 * Compute likelihood ratio distance netween two clusters.
 */
double cluster_bic_distance(cluster_t *c1, cluster_t *c2, float lambda)
{
  unsigned short dim = c1->dim;
  double d, p;  

  d = cluster_like_distance(c1, c2, 0);
  if (_ctype == CLUSTER_FULL_COV)
    p = (double)dim * (1.5 + 0.5 * (double)dim) * log(c1->nframes + c2->nframes);
  else
    p = 2.0 * (double)dim * log(c1->nframes + c2->nframes);

  return(d - 0.5 * lambda * p);
}

/* ------------------------------------------------------------------------------------ */
/* ----- double cluster_delta_bic(cluster_t *, cluster_t *, float, unsigned long) ----- */
/* ------------------------------------------------------------------------------------ */
/*
 * Return delta BIC for clustering of c1 and c2.
 */
double cluster_delta_bic(cluster_t *c1, cluster_t *c2, float lambda, unsigned long nframes)
{
  unsigned short dim = c1->dim;
  double d, p;

  d = cluster_like_distance(c1, c2, 0);
  if (_ctype == CLUSTER_FULL_COV)
    p = (double)dim * (1.5 + 0.5 * (double)dim) * log(nframes);
  else
    p = 2.0 * (double)dim * log(nframes);

  return(d - 0.5 * lambda * p);
}

#undef _cluster_c_

