/******************************************************************************/
/*                                                                            */
/*                                  estim.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:57 $
 * $Revision: 1.9 $
 *
 */

/*
 * Gaussian Mixture Modeling Parameter Estimation.
 */

#define _gmm_c_

#include <audioseg.h>

/* ------------------------------------------------------------------- */
/* ----- gmmacc_t *gmm_acc_alloc(unsigned short, unsigned short) ----- */
/* ------------------------------------------------------------------- */
/*
 * Allocate memory for the accumulators. Return a pointer to the
 * allocated accumulators or NULL in case of error.  
 */
gmmacc_t *gmm_acc_alloc(unsigned short n, unsigned short dim)
{
  gmmacc_t *p;
  unsigned short i;

  if ((p = (gmmacc_t *)malloc(sizeof(gmmacc_t))) == NULL)
    return(NULL);
  
  p->n = n;
  p->dim = dim; 
  p->w = NULL;
  p->m = NULL;
  p->v = NULL;

  if ((p->w = (double *)malloc(n * SIZEOF_DOUBLE)) == NULL) {
    gmm_acc_free(p);
    return(NULL);
  }
  
  if ((p->m = (double **)malloc(n * sizeof(double *))) == NULL) {
    gmm_acc_free(p);
    return(NULL);
  }
  for (i = 0; i < n; i++)
    *(p->m+i) = NULL;

  if ((p->v = (double **)malloc(n * sizeof(double *))) == NULL) {
    gmm_acc_free(p);
    return(NULL);
  }
  for (i = 0; i < n; i++)
    *(p->v+i) = NULL;

  for (i = 0; i < n; i++) {
    
    if ((*(p->m+i) = (double *)malloc(dim * SIZEOF_DOUBLE)) == NULL) {
      gmm_acc_free(p);
      return(NULL);
    }
    if ((*(p->v+i) = (double *)malloc(dim * SIZEOF_DOUBLE)) == NULL) {
      gmm_acc_free(p);
      return(NULL);
    }
  }

  return(p);
}

/* ------------------------------------------ */
/* ----- void gmm_acc_reset(gmmacc_t *) ----- */
/* ------------------------------------------ */
/*
 * Zero accumulator's values.
 */
void gmm_acc_reset(gmmacc_t *p)
{
  unsigned short i, j;
  double *wp, *mp, *vp;

  wp = get_acc_weights(p);

  for (i = 0; i < get_acc_num_comp(p); i++) {
    mp = get_acc_comp_mean(p, i);
    vp = get_acc_comp_var(p, i);

    *(wp+i) = 0.0;
    for (j = 0; j < get_acc_dim(p); j++) {
      *(mp+j) = 0.0;
      *(vp+j) = 0.0;
    }
  }
}

/* ----------------------------------------- */
/* ----- void gmm_acc_free(gmmacc_t *) ----- */
/* ----------------------------------------- */
/*
 * Free memory allocated to accumulator.
 */
void gmm_acc_free(gmmacc_t *p)
{
  unsigned short i;

  if (p) {

    if (p->w)
      free(p->w);

    if (p->m) {
      for (i = 0; i < p->n; i++)
        if (*(p->m+i))
          free(*(p->m+i));
      free(p->m);
    }
    
    if (p->v) {
      for (i = 0; i < p->n; i++)
        if (*(p->v+i))
          free(*(p->v+i));
      free(p->v);
    }

    free(p);
  }
}

/* ----------------------------------------------------------------------------------------- */
/* ----- void gmm_ml_update(gmm_t *, gmmacc_t *, unsigned long, int, double, double *) ----- */
/* ----------------------------------------------------------------------------------------- */
/*
 * Maximumu likelihood parameter re-estimation.
 */
void gmm_ml_update(gmm_t *gmm, gmmacc_t *accs, unsigned long nsamples, int update, double wfloor, double *vfloor)
{
  double *mp1, *vp1, *wp1; /* pointers in the accumulators */
  float *vp2, *mp2, *wp2;  /* pointers in the model */
  unsigned short i, j, dim = get_gmm_dim(gmm);
  unsigned short nfloored = 0;
  double x, z = (double)1.0 / (double)nsamples;

  wp1 = get_acc_weights(accs);
  wp2 = get_gmm_weights(gmm);

  /* compute new parameters in accumulators */
  for (i = 0; i < get_gmm_num_comp(gmm); i++) {
    mp1 = get_acc_comp_mean(accs, i);
    vp1 = get_acc_comp_var(accs, i);

    if (*(wp1+i) * z > GMM_MIN_WEIGHT) {
      x = 1.0 / *(wp1+i);
      *(wp1+i) *= z;

      for (j = 0; j < dim; j++) {
	*(mp1+j) *= x;
	*(vp1+j) = *(vp1+j) * x - *(mp1+j) * *(mp1+j);
      }
    }
    else {
      /* 
	 if not enough data for this component, kill it by setting the 
	 weight to zero and the mean and variance to dummy values.
      */
      *(wp1+i) = 0.0;
      for (j = 0; j < dim; j++) {
	*(mp1+j) = 0.0;
	*(vp1+j) = 1.0;
      }
    }
  }

  /* update model parameters */
  for (i = 0; i < get_gmm_num_comp(gmm); i++) {
    mp1 = get_acc_comp_mean(accs, i);
    vp1 = get_acc_comp_var(accs, i);
    mp2 = get_gmm_comp_mean(gmm, i);
    vp2 = get_gmm_comp_ivar(gmm, i);

    if (update & W_UP) {
      x = *(wp1+i);
      if (x > 0.0 && x < wfloor) {
        *(wp2+i) = (float)wfloor;
	nfloored++;
      } else
        *(wp2+i) = (float)x;
    }

    if (update & M_UP)
      for (j = 0; j < dim; j++)
        *(mp2+j) = (float)*(mp1+j);

    if (update & V_UP)
      for (j = 0; j < dim; j++) {
        x = *(vp1+j);
	if (vfloor)
	  *(vp2+j) = (x < *(vfloor+j)) ? (float)(1.0 / *(vfloor+j)) : ((float)(1.0 / x));
	else
	  *(vp2+j) = (float)(1.0 / x);
      }
  }
  
  /* renormalize weights, leaving floored weights unchanged */
  if (nfloored) { 
    x = 0;
    for (i = 0; i < get_gmm_num_comp(gmm); i++)
      if (*(wp2+i) > wfloor)
	x += *(wp2+i);

    x = (1.0 - nfloored * wfloor) / x;
    for (i = 0; i < get_gmm_num_comp(gmm); i++)
      if (*(wp2+i) > wfloor)
	*(wp2+i) *= x;
  }

  gmm_const_set(gmm);
}

/* ----------------------------------------------------------------------------------------------------------- */
/* ----- void gmm_map_update(gmm_t *, gmmacc_t *, unsigned long, int, double, double *, double, gmm_t *) ----- */
/* ----------------------------------------------------------------------------------------------------------- */
/*
 * Maximumu a posteriori parameter re-estimation.
 *
 * Details are given in D. Reynolds, T. Quatieri, and R. Dunn, "Speaker verification using 
 * adapted Gaussian mixture models", Digital Signal Processing Journal, 10(1-3):19-41, 2000.
 */
void gmm_map_update(gmm_t *gmm, gmmacc_t *accs, unsigned long nsamples, int update, double wfloor, double *vfloor, double r, gmm_t *prior)
{
  double *mp1, *vp1, *wp1; /* pointers in the accumulators */
  float *vp2, *mp2, *wp2;  /* pointers in the model */
  double sum, x, m, v;
  unsigned short i, j, nfloored = 0, dim = get_gmm_dim(gmm);
  double a, z = (double)1.0 / (double)nsamples;

  wp1 = get_acc_weights(accs);
  wp2 = get_gmm_weights(gmm);

  /* compute new parameters in accumulators */
  sum = 0;
  for (i = 0; i < get_gmm_num_comp(gmm); i++) {
    mp1 = get_acc_comp_mean(accs, i);
    vp1 = get_acc_comp_var(accs, i);
    mp2 =  get_gmm_comp_mean(prior, i);
    vp2 =  get_gmm_comp_ivar(prior, i);

    x = *(wp1+i);

    if (x * z > GMM_MIN_WEIGHT) {
      a = x / (r + x);

      *(wp1+i) = a * x * z + (1.0 - a) * get_gmm_comp_weight(prior, i);
      sum += *(wp1+i);

      x = 1.0 / x;
    
      for (j = 0; j < dim; j++) {
	m = *(mp2+j);
	v = 1.0 / *(vp2+j);

	*(mp1+j) = a * *(mp1+j) * x + (1.0 - a) * m;
	*(vp1+j) = a * *(vp1+j) * x + (1.0 - a) * (v + m * m) - *(mp1+j) * *(mp1+j);
      }
    }
    else {
      /* not enough data in this component -> copy prior model data */
      *(wp1+i) = get_gmm_comp_weight(prior, i);
      sum += *(wp1+i);
      for (j = 0; j < dim; j++) {
	*(mp1+j) =  (double)*(mp2+j);
	*(vp1+j) = (double)1.0 / (double)*(vp2+j);
      }
    }
  }

  sum = (double)1.0 / sum;

  /* update model parameters */
  for (i = 0; i < get_gmm_num_comp(gmm); i++) {
    mp1 = get_acc_comp_mean(accs, i);
    vp1 = get_acc_comp_var(accs, i);
    mp2 = get_gmm_comp_mean(gmm, i);
    vp2 = get_gmm_comp_ivar(gmm, i);

    if (update & W_UP) {
      x = *(wp1+i) * sum;
      if (x < wfloor) {
        *(wp2+i) = (float)wfloor;
	nfloored++;
      } else
        *(wp2+i) = (float)x;
    }

    if (update & M_UP)
      for (j = 0; j < dim; j++)
        *(mp2+j) = (float)*(mp1+j);


    if (update & V_UP)
      for (j = 0; j < dim; j++) {
        x = *(vp1+j);
	if (vfloor)
	  *(vp2+j) = (x < *(vfloor+j)) ? (float)(1.0 / *(vfloor+j)) : ((float)(1.0 / x));
	else
	  *(vp2+j) = (float)(1.0 / x);
      }
  }
  
  /* renormalize weights, leaving floored weights unchanged */
  if (nfloored) {
    x = 0;
    for (i = 0; i < get_gmm_num_comp(gmm); i++)
      if (*(wp2+i) > wfloor)
	x += *(wp2+i);

    x = (1.0 - nfloored * wfloor) / x;

    for (i = 0; i < get_gmm_num_comp(gmm); i++)
      if (*(wp2+i) > wfloor)
	*(wp2+i) *= x;
  }
  
  gmm_const_set(gmm);
}

/* --------------------------------------------------------------------- */
/* ----- double gmm_occ_probs(gmm_t *, spf_t *, float *, double *) ----- */
/* --------------------------------------------------------------------- */
/*
 * Compute the mixture component occupation probabilities for a
 * frame. Return the frame log-likelihood.
 */

#define EXP_MAX 700.0

double gmm_occ_probs(gmm_t *gmm, spf_t *yt, double *occprob)
{
  unsigned short i, j, ibest, dim = get_gmm_dim(gmm);
  static double s[GMM_NUM_COMP_MAX];
  double a, b, sum, sum1, sum2, logprob;
  float *w, *m, *v;
  unsigned short ncomp = get_gmm_num_comp(gmm);

  w = get_gmm_weights(gmm);
  ibest = 0;

  for (i = 0; i < ncomp; i++) {
    
    if (*(w+i) > GMM_MIN_WEIGHT) {
      sum = 0.0;
      m = get_gmm_comp_mean(gmm, i);
      v = get_gmm_comp_ivar(gmm, i);
    
      for (j = 0; j < dim; j++) {
	a = (double)*(yt+j) - (double)*(m+j);
	sum += a * a * (double)*(v+j);
      }

      *(s+i) = sum;
      if (sum < *(s+ibest))
	ibest = i;
    }
    else
      *(s+i) = 0.0;
  }

  /* compute log-likelihood */
  a = 1.0 / get_gmm_comp_expz(gmm, ibest);
  b = *(s+ibest);
  
  sum1 = get_gmm_comp_logc(gmm, ibest) - 0.5 * b;
  sum2 = 1.0;

  for (i = 0; i < ncomp; i++)
    if (i != ibest  &&  *(w+i) > GMM_MIN_WEIGHT)
      sum2 += a * get_gmm_comp_expz(gmm, i) * exp(-0.5 * (*(s+i) - b));

  logprob = sum1 + log(sum2);

  /* compute occupation probabilities */
  for (i = 0; i < ncomp; i++)
    if (*(w+i) > GMM_MIN_WEIGHT) {
      a = get_gmm_comp_logc(gmm, i) - 0.5 * *(s+i) - logprob;
      
      /* if the log prob is too small we can't compute the occprob */
      if (a > EXP_MAX)
	return DBL_MAX; 
      *(occprob+i) = exp(a);
    }
    else
      *(occprob+i) = 0.0;
  
  return(logprob);
}

/* -------------------------------------------------------------------------------------- */
/* ----- double gmm_accumulate(spfstream_t *, unsigned long *, gmm_t *, gmmacc_t *) ----- */
/* -------------------------------------------------------------------------------------- */
/*
 * Accumulate statistics for a segment. Return the segment total
 * log-likelihood and the actual number of frames used during the
 * accumulation. Note that the input feature stream must have been
 * seeked to the correct position before calling the accumulation
 * function.
 */
double gmm_accumulate(spfstream_t *f, unsigned long *nframes, gmm_t *gmm, gmmacc_t *accs)
{
  spf_t *yt;
  double logp, occprob[GMM_NUM_COMP_MAX];
  unsigned short i, j, dim, ncomp;
  unsigned long t, nused;
  double a, ytj, *w, *m, *v;

  ncomp = get_gmm_num_comp(gmm);
  dim = get_gmm_dim(gmm);
  w = get_acc_weights(accs);

  logp = 0;
  t = 0;
  nused = 0;

  while ((yt = get_next_spf_frame(f)) != NULL) {
 
    a = gmm_occ_probs(gmm, yt, occprob);

    if (a != DBL_MAX) {
      logp += a;
      
      nused++;
      for (i = 0; i < ncomp; i++) {
	m = get_acc_comp_mean(accs, i);
	v = get_acc_comp_var(accs, i);
	a = *(occprob+i);

	/* we don t need to do this if the gaussian is dead */
	if (a != 0.0) {
	  *(w+i) += a;
	
	  for (j = 0; j < dim; j++) {
	    ytj = *(yt+j);
	    *(m+j) += a * ytj;
	    *(v+j) += a * ytj * ytj;
	  }
	}
      }
    }

    t++;
    if (t == *nframes)
      break;
  }

  *nframes = nused;

  return(logp);
}

/* ---------------------------------------------------------------------- */
/* ----- double *set_var_floor(float, const char *, unsigned short) ----- */
/* ---------------------------------------------------------------------- */
/*
 * Return a variance floor vector. If a filename is specified, the
 * function reads variance from that ASCII file and sets the variance
 * floor to a fraction of the read variance. Otherwise, the variance
 * floor is set to the specified constant.  
 */
double *set_var_floor(float vfloor, const char *fn, unsigned short dim)
{
  FILE *f;
  double *p;
  unsigned short i;

  if ((p = (double *)malloc(dim * sizeof(double))) == NULL) {
    fprintf(stderr, "set_var_floor() -- cannot allocate memory\n");
    return(NULL);
  }

  if (fn) {
    if ((f = fopen(fn, "rt")) == NULL) {
      fprintf(stderr, "set_var_floor() -- cannot open variance file %s\n", fn);
      free(p);
      return(NULL);
    }
    for (i = 0; i < dim; i++)    
      if (fscanf(f, "%lf", p+i) != 1) {
	fprintf(stderr, "set_var_floor() -- cannot read variance bin %d in file %s\n", i+1, fn);
	free(p);
	return(NULL);	
      }
    fclose(f);

    for (i = 0; i < dim; i++)
      *(p+i) *= vfloor;
  }
  else
    for (i = 0; i < dim; i++)
      *(p+i) = vfloor;
  
  return(p);
}

#undef _gmm_c_
