/******************************************************************************/
/*                                                                            */
/*                                  hmm.c                                     */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Mike                                                              May 2003 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/08/25 13:32:45 $
 *   $Revision: 1.3 $
 *
 */

/*
 * HMM utilities.
 */

#define _hmm_c_

#include <hmm.h>

/* -------------------------------------------- */
/* ----- hmm_t *hmm_alloc(unsigned short) ----- */
/* -------------------------------------------- */
/*
 * Allocate memory for a HMM.
 */
hmm_t *hmm_alloc(unsigned short nstates)
{
  unsigned short i;
  hmm_t *p = (hmm_t *)malloc(sizeof(hmm_t));

  if (p) {
    p->nstates = nstates;
    p->state = NULL;
    p->entryp = NULL;
    p->transp = NULL;
    p->exitp = NULL;

    if ((p->state = (state_t *)malloc(nstates * sizeof(state_t))) == NULL) {
      free(p);
      return(NULL);
    }
    for (i = 0; i < nstates; i++) {
      p->state[i].name = NULL;
      p->state[i].pdf = NULL;
      p->state[i].complexity = 0.0;
    }

    if ((p->entryp = (float *)malloc(nstates * sizeof(float))) == NULL) {
      hmm_free(p);
      return(NULL);
    }

    if ((p->exitp = (float *)malloc(nstates * sizeof(float))) == NULL) {
      hmm_free(p);
      return(NULL);
    }

    if ((p->transp = (float **)malloc(nstates * sizeof(float *))) == NULL) {
      hmm_free(p);
      return(NULL);
    }
    for (i = 0; i < nstates; i++)
      p->transp[i] = NULL;
    for (i = 0; i < nstates; i++)
      if ((p->transp[i] = (float *)malloc(nstates * sizeof(float))) == NULL) {
	hmm_free(p);
	return(NULL);
      }
  }

  return(p);
}

/* ---------------------------------- */
/* ----- void hmm_free(hmm_t *) ----- */
/* ---------------------------------- */
/* 
 * Free memory allocated for a HMM.
 */
void hmm_free(hmm_t *p)
{
  unsigned short i;

  if (p) {
    if (p->state) {
      for (i = 0; i < p->nstates; i++) {
	if (p->state[i].name)
	  free(p->state[i].name);
	if (p->state[i].pdf)
	  gmm_free(p->state[i].pdf);
      }
      free(p->state);
    }
    
    if (p->entryp)
      free(p->entryp);
    
    if (p->exitp)
      free(p->exitp);
    
    if (p->transp) {
      for (i = 0; i < p->nstates; i++)
	if (p->transp[i])
	  free(p->transp[i]);
      free(p->transp);
    }
  }
}

/* ----------------------------------------- */
/* ----- hmm_t *hmm_read(const char *) ----- */
/* ----------------------------------------- */
/* 
 * Read a HMM from file.
 */
hmm_t *hmm_read(const char *fn)
{
  hmm_t *hmm = NULL;
  char line[HMM_MAX_LINE_LENGTH];
  char *pp, *p;
  unsigned short i, nstates, s = 0;
  int token = NUMBER_OF_STATES;     
  int lino = 0;
  float v;
  FILE *f;

  if ((f = fopen(fn, "r")) == NULL) {
    fprintf(stderr, "hmm_read() -- cannot open file %s\n", fn);
    return(NULL);
  }
  
  while (fgets(line, HMM_MAX_LINE_LENGTH, f) != NULL) {
    
    lino++;
    
    /* remove ending \n */
    if (line[strlen(line)-1] == '\n')
      line[strlen(line)-1] = 0x00; /* set end of string there */
    /* remove comments - i.e. search the # in the line */
    if ((p = (char *)strchr(line, HMM_COMMENT_CHAR)) != NULL)
      *p = 0x00; /* set end of string there */
    
    
    switch(token) {
      
    case NUMBER_OF_STATES:
      if (sscanf(line, "%hu", &nstates) != 1) {
	fprintf(stderr, "hmm_read() -- cannot read state number at line %d in file %s\n", lino, fn);
	fclose(f);
	return(NULL);
      }
      if ((hmm = hmm_alloc(nstates)) == NULL) {
	fprintf(stderr, "hmm_read() -- cannot create allocate memory\n");
	fclose(f);
	return(NULL);
      }
      token = STATE_DEFINITION;
      break;
      
    case STATE_DEFINITION:
      if ((p = strtok(line, " \t")) == NULL) {
	fprintf(stderr, "hmm_read() -- error reading state definition %d at line %d in file %s\n", s+1, lino, fn);
	hmm_free(hmm); fclose(f);
	return(NULL);
      }
      if ((hmm->state[s].name = strdup(p)) == NULL) {
	fprintf(stderr, "hmm_read() -- cannot allocate memory\n");
	hmm_free(hmm); fclose(f);
	return(NULL);
      }
      if ((p = strtok(NULL, " \t")) == NULL) {
	fprintf(stderr, "hmm_read() -- error reading state %d definition at line %d in file %s\n", s+1, lino, fn);
	hmm_free(hmm); fclose(f);
	return(NULL);
      }
      
      if ((hmm->state[s].pdf = gmm_read(p)) == NULL) {
	fprintf(stderr, "hmm_read() -- cannot load GMM from file %s (state %d, line %d, file %s)\n", p, s+1, lino, fn);
	hmm_free(hmm); fclose(f);
	return(NULL);
      }
      hmm->state[s].complexity =  0.5 * (get_gmm_num_comp(hmm->state[s].pdf) * (2.0 * get_gmm_dim(hmm->state[s].pdf) + 1) - 1);
      s++;
      if (s == nstates)
	token = STATE_ENTRY_PROB;
      break;

    case STATE_ENTRY_PROB:
      p = line;
      for (s = 0; s < nstates; s++) {
	v = strtod(p, &pp);
	if (! pp) {
	  fprintf(stderr, "hmm_read() -- error reading initial probabilities for state %d at line %d in file %s\n", s+1, lino, fn);	  
	  hmm_free(hmm); fclose(f);
	  return(NULL);
	}
	hmm->entryp[s] = (v == 0.0) ? -FLT_MAX : (float)log(v);
	p = pp;
      }
      token = STATE_TRANS_PROB;
      s = 0;
      break;
      
    case STATE_TRANS_PROB:
      p = line;
      for (i = 0; i < nstates; i++) {
	v = strtod(p, &pp);
	if (! pp) {
	  fprintf(stderr, "hmm_read() -- error reading transition probability a[%d][%d] at line %d in file %s\n",
		  s+1, i+1, lino, fn);
	  hmm_free(hmm); fclose(f);
	  return(NULL);
	}
	hmm->transp[s][i] = (v == 0.0) ? -FLT_MAX : (float)log(v);
	p = pp;
      }
      s++;
      if (s == nstates)
	token = STATE_EXIT_PROB;
      break;

    case STATE_EXIT_PROB:
      p = line;
      for (s = 0; s < nstates; s++) {
	v = strtod(p, &pp);
	if (! pp) {
	  fprintf(stderr, "hmm_read() -- error reading initial probabilities for state %d at line %d in file %s\n",
		  s+1, lino, fn);	  
	  hmm_free(hmm); fclose(f);
	  return(NULL);
	}
	hmm->exitp[s] = (v == 0.0) ? -FLT_MAX : (float)log(v);
	p = pp;
      }
      token = HMM_END;
      break;
      
    case HMM_END:
      break;

    default:
      return(NULL);
    }
  }
  
  fclose(f);
  
  return(hmm);
}
