/******************************************************************************/
/*                                                                            */
/*                                sbic.c                                      */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Mike                                                       Mon Aug  5 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) IRISA                                                       */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.12 $
 *
 */

/*
 * BIC segmentation.
 *
 * Segment the input file according to a Bayesian information
 * criterion (BIC), as defined in Perrine Delacourt, "La segmentation
 * et le regroupement par locuteurs pour l'indexation de documents
 * audio", Ph. D., Institu Eurecom. The approach implemented here is
 * somewhat different to the classical single-pass BIC segmentation
 * (see explanations below).
 *
 * If an input segmentation is given, the program will try to
 * resegment every input segment independently. One can also specify a
 * label so that every segment with the corresponding label will be
 * resegmented, the other segments being left untouched. This last
 * feature is particularly usefull in conjonction with ssad.
 *
 * The three pass algorithm works as follow: in the first pass, the
 * best change is searched in a window. If no change is detected in
 * that window, the window is extended. Otherwise, the change point is
 * memorized and we proceed with a new window starting where the
 * change was hypothesized. In the second pass, a thinner change
 * detection is performed with a window centered on each of the
 * changes hypothesized during the first pass. Finally, the third pass
 * consists in runing a BIC test between every pair of adjacent
 * segments.
 *
 * The single pass algorithm works with a sliding window and computes
 * the log-likelihood ratio. A very clever algorithm is then used to
 * find out important peaks in the likelihood ratio curve, such peaks
 * being considered as a change if above the threshold.
 *
 * NOTES: there may be some problems at the segment boundaries (such
 * as frames used twice).
 */

#define _sbic_c_

#include <audioseg.h>
#include <bic.h>
#include <getopt.h>

static char *cvsid = "$Id: sbic.c,v 1.12 2003/11/20 15:50:58 ggravier Exp $";

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
unsigned long inc1 = 60;       /* 1st pass increment                   */
unsigned long size1 = 300;     /* 1st pass original window size        */
unsigned long inc2 = 20;       /* 2nd pass increment                   */
unsigned long size2 = 200;     /* 2nd pass original window size        */
float lambda = 1.5;            /* model complexity penalty             */
int fullc = 0;                 /* use full covariance matrices         */
char *segfn = NULL;            /* segmentation filename                */
char *labname = NULL;          /* extract label name                   */
float t_start = ASEG_NULL_TIME;/* start time                           */
float t_end = ASEG_NULL_TIME;  /* end time                             */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int help = 0;                  /* print usage                          */
int trace = 0;                 /* trace level                          */
int version = 0;               /* version flag                         */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sbic [options] ifile ofile\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  BIC segmentation. If the second pass window size is not null, a three\n");
  fprintf(stdout, "  pass segmentation scheme is used. Otherwise, a single pass approach is\n");
  fprintf(stdout, "  used with lambda as the decision threshold.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -b, --size-1=n               first pass window size (%lu samples)\n", size1);
  fprintf(stdout, "  -a, --increment-1=n          first pass increment (%lu samples)\n", inc1);
  fprintf(stdout, "  -d, --size-2=n               second pass window size (%lu samples)\n", size2);
  fprintf(stdout, "  -c, --increment-2=n          second pass increment (%lu samples)\n", inc2);
  fprintf(stdout, "  -l, --lambda=f               complexity penalty weight (%.2f)\n", lambda);
#if HAVE_GSL
  fprintf(stdout, "  -f, --full                   use full covariance matrices (diagonal)\n");
#endif
  fprintf(stdout, "\n");
  fprintf(stdout, "  -s, --segmentation=fn        use segmentation (don't)\n");
  fprintf(stdout, "  -x, --label=s                extract label from segmentation (none)\n");
  fprintf(stdout, "  -t, --time=[st][:et]         specifiy start and end times (whole file)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ZRADN (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);  
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char** argv)
{
  char *ifn, *ofn;                /* I/O filenames                            */
  spfstream_t *f;                 /* input feature stream                     */
  asseg_t *iseg = NULL;           /* input segmentation if any                */
  seglab_t *label = NULL;         /* input label if any                       */
  asseg_t *oseg;                  /* BIC segmentation                         */

  int read_args(int, char **);
  asseg_t *bic(spfstream_t *, asseg_t *, seglab_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(1);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }

  /* read arguments */
  if (optind < argc)
    ifn = argv[optind++];
  else {
    fprintf(stderr, "sbic error -- no input filename specified (use --help to get usage)\n");
    return(1);
  }
  
  if (optind < argc)
    ofn = argv[optind++];
  else {
    fprintf(stderr, "sbic error -- no output filename specified (use --help to get usage)\n");
    return(1);
  }
 
  if (optind < argc) {
    fprintf(stderr, "sbic error -- invalid number of arguments (use --help to get usage)\n");
    return(1);
  }
  
  if (trace) {
    fprintf(stdout, "segmenting file %s\n", ifn ? ifn : "stdin");
    fflush(stdout);
  }

  /* ----- open input stream and segmentation ----- */
  if ((f = spf_input_stream_open(ifn, cvtflag, iobs)) == NULL) {
    fprintf(stderr, "sbic error -- cannot open feature stream %s\n", ifn ? ifn : "stdin");
    return(1);
  }
  
  if (segfn)
    if ((iseg = seg_read(segfn)) == NULL) {
      fprintf(stderr,"sbic error -- cannot load segmenation file %s\n", segfn);
      spf_stream_close(f);
      return(0);
    }
  
  /* ----- create label if necessary ----- */
  if (labname)
    if ((label = str_to_label(labname)) == NULL) {
      fprintf(stderr, "sbic error -- cannot create reference label\n");
      spf_stream_close(f); seg_list_free(iseg);
      return(1);
    }

  /* ----- run segmentation ----- */
  if ((oseg = bic(f, iseg, label)) == NULL) {
    spf_stream_close(f); seg_list_free(iseg); seg_label_free(label);
    return(1);
  }

  /* ---- write output segmentation ---- */
  if (trace) {
    fprintf(stdout, "writing segmentation to file %s\n", ofn ? ofn : "stdout");
    fflush(stdout);
  }
  if (seg_write(oseg, ofn, "lse") == 0) {
    fprintf(stderr, "sbic error -- unable to write segment list to file %s\n", ofn ? ofn : "stdout");
    spf_stream_close(f); seg_list_free(iseg); seg_list_free(oseg); seg_label_free(label);
    return(1);
  }

  /* ---- the end ---- */
  seg_list_free(iseg);
  seg_label_free(label);
  spf_stream_close(f);
  seg_list_free(oseg);

  return(0);
}



/* -------------------------------------------------------------- */
/* ----- asseg_t *bic(spfstream_t *, asseg_t *, seglab_t *) ----- */
/* -------------------------------------------------------------- */
/*
 * BIC segmentation.
 */
asseg_t *bic(spfstream_t *f, asseg_t *iseg, seglab_t *label)
{
  asseg_t *curseg, *p, *ip;
  asseg_t *seg;                   /* output segmentation                      */
  
  asseg_t *bic_segment(spfstream_t *, float, float); /* three pass BIC segmenter */
  asseg_t *seg_copy(asseg_t *);
  
  if (iseg == NULL) {
    if ((seg = bic_segment(f, t_start, t_end)) == NULL) {
      fprintf(stderr, "sbic error -- cannot run BIC segmentation on input stream\n");
      return(NULL);
    }
  }
  else {
    seg = curseg = NULL;
    ip = iseg;

    while (ip) {
      if (seg_label_compare(label, get_seg_label(ip)) == 0) {

	/* run BIC segmentation and append the result to output segmentation */
	if ((p = bic_segment(f, get_seg_start_time(ip), get_seg_end_time(ip))) == NULL) {
	  fprintf(stderr, "sbic error -- cannot run BIC segmentation on current segment\n");
	  seg_list_free(seg);
	  return(NULL);
	}

	p->prev = curseg;
	if (curseg)
	  curseg->next = p;
	else
	  seg = p;
	
	/* move curseg to the last element of p */
	curseg = p;
	while (curseg->next)
	  curseg = curseg->next;
      }
      else { 

	/* copy and append to output segmentation */
	if ((p = seg_copy(ip)) == NULL) {
	  fprintf(stderr, "sbic error -- cannot allocate memory\n");
	  seg_list_free(seg);
	  return(NULL);
	}

	p->prev = curseg; 
	if (curseg)
	  curseg->next = p;
	else
	  seg = p;

	/* move curseg to the last element of p */
	curseg = p;
      }

      ip = ip->next;
    }
  }

  return(seg);
}

/* ------------------------------------------------------------- */
/* ----- asseg_t *bic_segment(spfstream_t *, float, float) ----- */
/* ------------------------------------------------------------- */
/*
 * Three pass BIC segmenter on the specified portion. Returns the
 * corresponding segmentation.
 * 
 */
asseg_t *bic_segment(spfstream_t *f, float st, float et)
{
  asseg_t *one_pass_bic(spfstream_t *, float, float);
  asseg_t *three_pass_bic(spfstream_t *, float, float);

  if (size2)
    return(three_pass_bic(f, st, et));
  
  return(one_pass_bic(f, st, et));
}

/* -------------------------------------------------------------- */
/* ----- asseg_t *one_pass_bic(spfstream_t *, float, float) ----- */
/* -------------------------------------------------------------- */
/*
 * Three pass BIC segmenter on the specified portion. Returns the
 * corresponding segmentation.
 * 
 */
asseg_t *one_pass_bic(spfstream_t *f, float st, float et)
{
  bicwin_t *win;                  /* window for segmentation                  */
  unsigned long sn, en, nsegs = 0;
  unsigned long i, j, m, n, midstep = size1 / 2;
  asseg_t *seg, *curseg, *p;
  char name[] = BIC_OUTPUT_STRING;  /* output segment label                     */
  double z = (double)1.0 / (double)spf_stream_rate(f);

  struct {
    double d;
    float t;
  } *tab;
  double d, d1, d2;

  sn = (st == ASEG_NULL_TIME) ? (unsigned long)0 : (unsigned long)(st * spf_stream_rate(f));
  en = (et == ASEG_NULL_TIME) ? (unsigned long)0 : (unsigned long)(et * spf_stream_rate(f));

  if ((tab = malloc(1000 * sizeof(*tab))) == NULL) {
    fprintf(stderr, "sbic error -- cannot allocate memory\n");
    return(NULL);
  }
  m = 1000;

  if ((win = bic_init(f, sn, size1, en, fullc)) == NULL) {
    fprintf(stderr, "sbic error -- error initializing BIC window\n");
    free(tab);
    return(NULL);
  }
  
  seg = curseg = NULL;

  /* compute distance */
  if (trace > 2) {
    fprintf(stdout, "computing BIC distances\n");
    fflush(stdout);
  }

  n = 0;
  while (win->buf->n) {

    if (n == m) {
      if ((tab = realloc(tab, (m + 1000) * sizeof(*tab))) == NULL) {
	fprintf(stderr, "sbic error -- cannot allocate memory\n");
	return(NULL);
      }
      m += 1000;
    }

    win->im = win->is + midstep;
    tab[n].d = -dbic(win, 0);
    tab[n].t = win->im * z;

    n++;

    if (bic_win_fill(win, win->is + inc1, size1) < size1)
      break;
  }

  /* compute emergence from BIC distances */
  if (trace > 2) {
    fprintf(stdout, "searching changes\n");
    fflush(stdout);
  }

  if ((seg = seg_create(name, (st == ASEG_NULL_TIME) ? 0.0 : st, ASEG_NULL_TIME, 0.0)) == NULL) {
    fprintf(stderr, "sbic error -- cannot allocate memory\n");
    bic_reset(win); free(tab);
    return(NULL);
  }
  curseg = seg;
  nsegs = 1;
  
  for (i = 0; i < n; i++) {
    
    d1 = DBL_MAX;
    for (j = i; j > 0; j--) {
      d = tab[j-1].d;
      if (d >= tab[i].d)
	break;
      if (d < d1)
	d1 = d;
    }

    d2 = DBL_MAX;
    for (j = i + 1; j < n; j++) {
      d = tab[j].d;
      if (d >= tab[i].d)
	break;
      if (d < d2)
	d2 = d;
    }

    /* if we are not at a local maximum or if we are at the boundary,
       no change is possible. */
    if (d1 == DBL_MAX || d2 == DBL_MAX) 
      d = 0;
    else {
    
      if (d1 == DBL_MAX)
	d1 = 0;
      if (d2 == DBL_MAX)
	d2 = 0;
      d = (d1 > d2) ? tab[i].d - d1 : tab[i].d - d2;
    }

    if (d > lambda) { /* found a segment */

      if (trace > 2)
	fprintf(stdout, "  change found at t=%.2f\n", tab[i].t);

      curseg->et = tab[i].t;

      if ((p = seg_create(name, tab[i].t, ASEG_NULL_TIME, 0.0)) == NULL) {
	fprintf(stderr, "sbic error -- cannot allocate memory\n");
	bic_reset(win); seg_list_free(seg); free(tab);
	return(NULL);
      }

      curseg->next = p;
      p->prev = curseg;
      curseg = p;
      nsegs++;
    }
  }
  curseg->et = win->ie * z;
  
  if (trace > 1) {
    fprintf(stdout, "first pass segmentation: %lu segment(s) found\n", nsegs);
    fflush(stdout);
  }

  bic_reset(win);
  free(tab);

  return(seg);
}

/* ---------------------------------------------------------------- */
/* ----- asseg_t *three_pass_bic(spfstream_t *, float, float) ----- */
/* ---------------------------------------------------------------- */
/*
 * Three pass BIC segmenter on the specified portion. Returns the
 * corresponding segmentation.
 * 
 */
asseg_t *three_pass_bic(spfstream_t *f, float st, float et)
{
  bicwin_t *win;                  /* window for segmentation                  */
  unsigned long sn, en, nsegs = 0;
  unsigned long i, n;
  asseg_t *seg, *curseg, *p;
  char name[] = BIC_OUTPUT_STRING;  /* output segment label                     */
  double z = (double)1.0 / (double)spf_stream_rate(f);

  sn = (st == ASEG_NULL_TIME) ? (unsigned long)0 : (unsigned long)(st * spf_stream_rate(f));
  en = (et == ASEG_NULL_TIME) ? (unsigned long)0 : (unsigned long)(et * spf_stream_rate(f));

  if ((win = bic_init(f, sn, size1, en, fullc)) == NULL) {
    fprintf(stderr, "sbic error -- error initializing BIC window\n");
    return(NULL);
  }

  seg = curseg = NULL;

  /*
   * first pass -- rough change detection
   */
  if ((seg = seg_create(name, sn, 0.0, 0.0)) == NULL) {
    fprintf(stderr, "sbic error -- cannot allocate memory\n");
    bic_reset(win);
    return(NULL);
  }
  curseg = seg;
  nsegs = 1;

  if (trace > 2) {
    fprintf(stdout, "first pass:\n");
    fflush(stdout);
  }

  while (win->buf->n) {
    if (bic_change_search(win, inc1, lambda, &i)) {

      /* found a change at index i in window : start a new segment
	 with starting time i and restart from current index in the
	 input stream */
      nsegs++;

      if (trace > 2)
	fprintf(stdout, "  change hpothesized at t=%.2f\n", i * z);
      
      curseg->et = i;

      if ((p = seg_create(name, i, 0.0, 0.0)) == NULL) {
	fprintf(stderr, "sbic error -- cannot allocate memory\n");
	bic_reset(win); seg_list_free(seg);
	return(NULL);
      }

      curseg->next = p;
      p->prev = curseg;
      curseg = p;

      /* update window for next search */
      if (bic_win_fill(win, i + inc1, size1) < 2 * inc1)
	break;
    }
    else {
      n = win->buf->n;
      if (bic_win_fill(win, win->is, n + size1) < n + 2 * inc1)
	break;
    }
  }
  curseg->et = win->ie;

  if (trace > 1) {
    fprintf(stdout, "first pass segmentation: %lu segment(s) found\n", nsegs);
    fflush(stdout);
  }

  /*
   * second pass -- thinner change detection 
   */

  if (trace > 2) {
    fprintf(stdout, "second pass:\n");
    fflush(stdout);
  }

  curseg = seg->next;
  nsegs = 1;
  while (curseg) {

    if (curseg->st > size2 / 2)
      sn = (unsigned long)(curseg->st) - size2 / 2;
    else
      sn = 0;

    if (bic_win_fill(win, sn, size2) == 0) {
      fprintf(stderr, "sbic error -- cannot fill in BIC segmentation buffer\n");
      bic_reset(win); seg_list_free(seg);
      return(NULL);
    }
    
    if (bic_change_search(win, inc2, lambda, &i) && i > curseg->prev->st && i < curseg->et) {

      if (curseg->st != (float)i) {
	/* move change to new index i */
	if (trace > 2)
	  fprintf(stdout, "  change moved from t=%.2f to t=%.2f\n", curseg->st * z, i * z);

	curseg->st = curseg->prev->et = i;
      }
      nsegs++;
    }
    else {
      /* suppress change */

      if (trace > 2)
	fprintf(stdout, "  change suppressed at t=%.2f\n", curseg->st * z);

      p = curseg->prev;

      p->et = curseg->et;
      p->next = curseg->next;
      if (curseg->next)
	curseg->next->prev = p;
      seg_free(curseg);
      curseg = p;
    }
    
    curseg = curseg->next;
  }

  if (trace > 1) {
    fprintf(stdout, "second pass segmentation: %lu segment(s) found\n", nsegs);
    fflush(stdout);
  }

  /*
   * third pass -- change validation 
   */

  if (trace > 2) {
    fprintf(stdout, "third pass:\n");
    fflush(stdout);
  }

  curseg = seg;

  while (curseg->next) {
    
    if (bic_win_fill(win, curseg->st, curseg->next->et - curseg->st) == 0) {
      fprintf(stderr, "sbic error -- cannot fill in BIC segmentation buffer\n");
      bic_reset(win); seg_list_free(seg);
      return(NULL);
    }
    win->im = curseg->et;

    if (dbic(win, lambda) > 0) {

      if (trace > 2)
	fprintf(stdout, "  change suppressed at t=%.2f\n", curseg->et * z);

      p = curseg->next;
      curseg->et = p->et;
      curseg->next = p->next;
      if (p->next)
	p->next->prev = curseg;
      seg_free(p);

      nsegs--;
    }
    else
      curseg = curseg->next;
  }

  if (trace > 1) {
    fprintf(stdout, "third pass segmentation: %lu segment(s) found\n", nsegs);
    fflush(stdout);
    }

  /*  
   * last but not least, convert absolute frame indexes into times.  
   */
  curseg = seg;
  while (curseg) {
    curseg->st *= z;
    curseg->et *= z;
    curseg = curseg->next;
  }

  bic_reset(win);

  return(seg);
}

/* ---------------------------------------- */
/* ----- asseg_t *seg_copy(asseg_t *) ----- */
/* ---------------------------------------- */
asseg_t *seg_copy(asseg_t *p)
{
  unsigned short i;
  asseg_t *pp;

  if ((pp = seg_alloc()) == NULL)
    return(NULL);

  pp->st = p->st;
  pp->et = p->et;
  pp->score = p->score;
  pp->next = p->next;
  pp->prev = p->prev;

  if ((pp->label = seg_label_alloc(p->label->nlabels)) == NULL) {
    seg_free(pp);
    return(NULL);
  }

  for (i = 0; i < p->label->nlabels; i++)
    if ((pp->label->name[i] = strdup(p->label->name[i])) == NULL) {
      seg_free(pp);
      return(NULL);
    }
  
  return(pp);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -b, --size-1=f              first pass window size
 *  -a, --increment-1=f         first pass increment
 *  -d, --size-2=f              second pass window size
 *  -c, --increment-2=f         second pass increment
 *  -l, --lambda                complexity penalty weight
 *  -f, --full                  use full covariance matrices
 *  -s, --segmentation=fn       resegment according to segmentation
 *  -x, --label=s               resegment segments with specified name
 *  -t, --time=[st][:et]         specifiy start and end times
 *  -C, --convert=s             convertion string - ADNZ
 *  -S, --buffer-size=n         input buffer size in kbytes
 *  -v, --verbose[=n]           verbose mode
 *  -V, --version               print version number
 *  -h, --help                  this help message
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"convert", required_argument, NULL, 'C'},
    {"increment-1", required_argument, NULL, 'a'},
    {"size-1", required_argument, NULL, 'b'},
    {"increment-2", required_argument, NULL, 'c'},
    {"size-2", required_argument, NULL, 'd'},
    {"lambda", required_argument, NULL, 'l'},
#if HAVE_GSL
    {"full", no_argument, NULL, 'f'},
#endif
    {"time", required_argument, NULL, 't'},
    {"segmentation", required_argument, NULL, 's'},
    {"label", required_argument, NULL, 'x'},    
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},    
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

#if HAVE_GSL
  while ((c = getopt_long(argc, argv, "C:a:b:c:d:l:fv::t:s:x:S:Vh", longopts, &i)) != EOF)
#else
  while ((c = getopt_long(argc, argv, "C:a:b:c:d:l:v::t:s:x:S:Vh", longopts, &i)) != EOF)
#endif
    switch (c) {

    case 'a':
      inc1 = strtoul(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sbic error -- invalid argument %s to --first-increment (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'b':
      size1 = strtoul(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sbic error -- invalid argument %s to --first-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'c':
      inc2 = strtoul(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sbic error -- invalid argument %s to --second-increment (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'd':
      size2 = strtoul(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sbic error -- invalid argument %s to --second-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'l':
      lambda = strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sbic error -- invalid argument %s to --lambda (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
#if HAVE_GSL
    case 'f':
      fullc = 1;
      break;
#endif

    case 's':
      segfn = optarg;
      break;
      
    case 'x':
      labname = optarg;
      break;

    case 't':
      p = optarg;
      if (*p != ':') {
	t_start = (float)strtod(optarg, &p);
	if (p == optarg || (*p != 0 && *p != ':')) {
	  fprintf(stderr, "sbic error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      if (*p == ':')
	t_end = (float)strtod(p+1, &p);
      
      if (*p) {
	fprintf(stderr, "sbic error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;

    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
        fprintf(stderr, "sbic error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
        return(1);
      }
      break;

    case 'v':
      if (optarg) {
        trace = (int)strtol(optarg, &p, 0);
        if (p == optarg || *p != 0) {
          fprintf(stderr, "sbic error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
          return(1);
        }
      }
      else
	trace = 1;
      break;

    case 'V':
      version = 1;
      return(0);
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sbic error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  /* argument sanity check */
  if (inc1 == 0 || size1 == 0) {
    fprintf(stderr, "sbic error -- invalid first pass increment %lu or window size %lu (both must be positive)\n", inc1, size1);
    return(1);
  }

  if (inc2 == 0) {
    fprintf(stderr, "sbic error -- invalid second pass increment %lu or window size %lu (both must be positive)\n", inc2, size2);
    return(1);
  }

  if (segfn && (t_start != ASEG_NULL_TIME || t_end != ASEG_NULL_TIME)) {
    fprintf(stderr, "sbic error -- cannot specify segmentation and start or end times (choose one)\n");
    return(1);
  }

  return(0);
}

#undef _sbic_c_
