/******************************************************************************/
/*                                                                            */
/*                               scluster.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/08/22 14:15:11 $
 * $Revision: 1.2 $
 *
 */

/*
 * BIC clustering.
 *
 * Cluster the input segments according to the Bayesian information
 * criterion (BIC) and a similarity measure between two clusters. The
 * similarity measure is one of:
 *  - symmetric Kullback-Liebler,
 *  - log-likelihood ratio, or 
 *  - delta BIC
 *
 * For details, see Scott S. Chen and P. S. Gopalakrishnan, "Speaker,
 * environment and channel change detection and clustering via the
 * Bayesian information criterion", in Proc. of the Broadcast News
 * Transcription and Understanding Workshop, pp. 127--132, 1998.  An
 * iterative variant is also implemented here.
 *
 * The program can operate either on a single data file and the
 * corresponding segmentation or on a script. In this last case, the
 * script must contain segmentation files. Bottom-up clustering is
 * performed on all the input segments unless a label is specified, in
 * which case only the segments with that particular label are
 * clustered, the remaining segments being left unchanged.
 */

#define _scluster_c_

#include <audioseg.h>
#include <script.h>
#include <cluster.h>
#include <getopt.h>

static char *cvsid = "$Id: scluster.c,v 1.2 2003/08/22 14:15:11 ggravier Exp $";

# define MAX_FILE_NAME_LENGTH 4096

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
char *scpfn = NULL;            /* script filename                      */
char *labname = NULL;          /* extract label name                   */
unsigned long mclusters = 1;   /* minimum number of clusters           */
double maxd = DBL_MAX;         /* max. distance                        */
float lambda = 1.5;            /* complexity penalty weight            */
int kl2 = 0;                   /* use symmetric Kullback-Liebler       */    
int ratio = 0;                 /* use log-likelihood ratio distance    */
int fullc = 0;                 /* use full covariance matrices         */
int iterative = 0;             /* run in iterative mode                */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
char *suffix = NULL;           /* output file suffix                   */
int merge = 0;                 /* merge adjacent segments              */
int help = 0;                  /* print usage                          */
int trace = 0;                 /* trace level                          */
int version = 0;               /* print version information            */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  scluster [options] data seg [oseg]\n");
  fprintf(stdout, "  scluster [options] --file-list=fn\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  BIC-based bottom-up clustering\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -k, --kullback               use symmetric Kullback-Liebler distance (BIC)\n");
  fprintf(stdout, "  -r, --ratio                  use log likelihood ratio distance (BIC)\n");
  fprintf(stdout, "  -l, --lambda=f               complexity penalty weight (%.2f)\n", lambda);
#if HAVE_GSL
  fprintf(stdout, "  -f, --full                   use full covariance matrices (diagonal)\n");
#endif
  fprintf(stdout, "\n");
  fprintf(stdout, "  -d, --max-distance=f         distance above which clusters are not merged (inf)\n");
  fprintf(stdout, "  -n, --num-clusters=n         minimum number of clusters (%lu)\n", mclusters);
  fprintf(stdout, "  -i, --iterate                iterative mode (no)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -L, --file-list=fn           input script (none)\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ADNZ (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);  
  fprintf(stdout, "\n");
  fprintf(stdout, "  -x, --label=s                score segments with specified name (none)\n");
  fprintf(stdout, "  -b, --suffix=fn              output suffix (in place)\n");
  fprintf(stdout, "  -m, --merge                  merge adjacent segments in same cluster (don't)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char** argv)
{
  char *ofn = NULL;
  script_t *script;
  seglab_t *label = NULL;

  int read_args(int, char **);
  int bic_clustering(script_t *, seglab_t *);
  unsigned long write_script_seg(script_t *, const char *, const char *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) { 
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }
  
  /* ----- load script ----- */
  if (scpfn) {
    if ((script = load_script(scpfn, 0)) == NULL) {
      fprintf(stderr, "scluster error -- cannot load script from file %s\n", scpfn);
      return(1);
    }
  }
  else {
    if (optind >= argc) {
      fprintf(stderr, "scluster error -- no data filename specified (use --help to get usage)\n");
      return(1);
    }

    if ((script = make_script(1, argv + optind)) == NULL) {
      fprintf(stderr, "scluster error -- cannot create script from arguments\n");
      return(1);
    }

    if (++optind >= argc) {
      fprintf(stderr, "scluster error -- no segmentation filename specified (use --help to get usage)\n");
      return(1);
    }

    if ((script->segfn[0] = strdup(argv[optind++])) == NULL) {
      fprintf(stderr, "scluster error -- cannot allocate memory\n");
      free_script(script);
      return(1);
    }

    if (optind < argc)
      ofn = argv[optind++];

    if (optind < argc) {
      fprintf(stderr, "scluster error -- two many arguments (use --help to get usage)\n");
      free_script(script);
      return(1);
    }
  }

  /* ----- load segmentations ----- */
  if (load_script_seg(script) == NULL) {
    fprintf(stderr, "scluster error -- cannot load segmentation(s)\n");
    free_script(script);
    return(1);
  }

  /* ----- create label if necessary ----- */
  if (labname)
    if ((label = str_to_label(labname)) == NULL) {
      fprintf(stderr, "scluster error -- cannot create reference label\n");
      free_script(script);
      return(1);
    }
  
  /* ----- run clustering ----- */
  if (bic_clustering(script, label) != 0) {
    free_script(script); seg_label_free(label);
    return(1);    
  }

  /* ----- output new segmentation(s) ----- */
  if (write_script_seg(script, ofn, suffix) != script->nfiles) {
    free_script(script); seg_label_free(label);
    return(1);
  }

  /* ----- clean and go to bed ----- */
  free_script(script);
  seg_label_free(label);

  return(0);
}

/* ------------------------------------------------------ */
/* ----- int bic_clustering(script_t *, seglab_t *) ----- */
/* ------------------------------------------------------ */
/*
 * Perform BIC clustering on the segments in script corresponding to
 * the specified labels, or all the segments if no label. Script
 * segmentations are modified in place after clusters have been
 * determined.  
 */
int bic_clustering(script_t *script, seglab_t *label)
{
  cluster_t *clusters;
  unsigned long nframes, nclusters;

  cluster_t *init_clusters(script_t *, seglab_t *, unsigned long *, unsigned long *);
  cluster_t *build_clusters(cluster_t *, unsigned long *, unsigned long);
  int make_segmentations(script_t *, cluster_t *);

  /* ----- initialize clusters with segments ----- */

  if ((clusters = init_clusters(script, label, &nframes, &nclusters)) == NULL) {
    fprintf(stderr, "scluster error -- cannot initialize clusters from script\n");
    return(1);
  }

  if (nclusters < 2) {
    fprintf(stderr, ">>>>> WARNING >>>>>> no or single initial cluster (maybe a bad label name)\n");
    return(0);
  }

  if (trace > 1) {
    fprintf(stdout, "number of segments = %lu\n", nclusters);
    fprintf(stdout, "total length = %lu frames\n", nframes);
    fflush(stdout);
  }

  /* ----- perform clustering ----- */
  if ((clusters = build_clusters(clusters, &nclusters, nframes)) == NULL) {
    cluster_reset();
    return(1);
  }

  if (trace > 1) {
    fprintf(stdout, "number of clusters = %lu\n", nclusters);
    fprintf(stdout, "saving segmentations to disk:\n");
    fflush(stdout);
  }

  if (make_segmentations(script, clusters) != 0) {
    cluster_reset(); cluster_list_free(clusters);
    return(1);
  }

  cluster_reset();

  return(0);
}

/* ----------------------------------------------------------------------------------------------- */
/* -----  cluster_t *init_clusters(script_t *, seglab_t *, unsigned long *, unsigned long *) ----- */
/* ----------------------------------------------------------------------------------------------- */
/*
 * Initialize clusters from segments.
 */
cluster_t *init_clusters(script_t *script, seglab_t *label, unsigned long *nframes, unsigned long *nclusters)
{
  unsigned short dim = 0; /* makes gcc happy! */
  unsigned long i;
  spfstream_t *f;
  asseg_t *seg;
  cluster_t *p, *pp, *listp;
  cluster_t *init_cluster(spfstream_t *, float, float);

  *nclusters = 0;
  *nframes = 0;
  pp = NULL;
  listp = NULL;

  for (i = 0; i < script->nfiles; i++) {

    if ((f = spf_input_stream_open(script->datfn[i], cvtflag, iobs)) == NULL) {
      fprintf(stderr, "scluster error -- cannot open input stream %s\n", script->datfn[i]);
      cluster_list_free(listp);
      return(NULL);
    }

    if (i) {
      if (spf_stream_dim(f) != dim) {
	fprintf(stderr, "scluster error -- dimension %hu in %s incompatible with %hu\n", spf_stream_dim(f), script->datfn[i], dim);
	cluster_reset(); spf_stream_close(f); cluster_list_free(listp);
	return(NULL);
      }
    }
    else {
      dim = spf_stream_dim(f);
      if (cluster_init(dim, fullc) != 0) {
	fprintf(stderr, "scluster error -- cannot allocate memory\n");
	spf_stream_close(f);
	return(NULL);
      }
    }

    seg = script->seg[i];
    while (seg) {

      if (seg_label_compare(get_seg_label(seg), label) == 0) {

	if ((p = init_cluster(f, get_seg_start_time(seg), get_seg_end_time(seg))) == NULL) {
	  fprintf(stderr, "scluster error -- cannot create new cluster\n");
	  cluster_reset(); spf_stream_close(f); cluster_list_free(listp);
	  return(NULL);
	}

	if ((p->link = (seglink_t *)malloc(sizeof(seglink_t))) == NULL) {
	  fprintf(stderr, "scluster error -- cannot allocate memory\n");
	  cluster_reset(); spf_stream_close(f); cluster_list_free(listp);
	  return(NULL);
	}
	p->link->seg = seg;
	p->link->next = NULL;


	p->prev = pp;
	if (pp)
	  pp->next = p;
	else
	  listp = p;
	pp = p;

	*nclusters += 1;
	*nframes += p->nframes;
      }
      
      seg = get_seg_next(seg);
    }

    spf_stream_close(f);
  }

  return(listp);
}

/* ---------------------------------------------------------------- */
/* ----- cluster_t *init_cluster(spfstream_t *, float, float) ----- */
/* ---------------------------------------------------------------- */
/*
 * Initialize a cluster from a feature stream.
 */
cluster_t *init_cluster(spfstream_t *f, float st, float et)
{
  unsigned long t, sn, en, nframes;
  unsigned short i, j, dim = spf_stream_dim(f);
  cluster_t *p;
  double *m, *v;
  spf_t *yt, x;

  /* get memory */
  if ((p = cluster_alloc(dim)) == NULL) {
    fprintf(stderr, "scluster error -- cannot allocate memory\n");
    return(NULL);
  }

  /* seek into stream */
  sn = (st == ASEG_NULL_TIME) ? 0 : (unsigned long)(st * spf_stream_rate(f));
  if (et != ASEG_NULL_TIME) {
    en = (unsigned long)(et * spf_stream_rate(f));
    if (en == sn)
      en += 1;
  }
  else
    en = 0;

  if (spf_stream_seek(f, sn, SEEK_SET) != 0) {
    fprintf(stderr, "scluster error -- cannot seek frame %lu at t=%.f\n", sn, st);
    cluster_free(p);
    return(NULL);
  }

  t = sn;
  nframes = 0;
  m = p->m;
  v = *(p->v);

  for (i = 0; i < dim; i++) {
    *(m+i) = 0.0;

    if (fullc) {
      v = *(p->v+i);
      for (j = 0; j < dim - i; j++)
	*(v+j) = 0.0;
    }
    else 
      *(v+i) = 0.0;
  }

  while ((yt = get_next_spf_frame(f)) != NULL) {

    for (i = 0; i < dim; i++) {
      x = *(yt+i);
      *(m+i) += x;
      
      if (fullc) {
	v = *(p->v+i);
	for (j = 0; j < dim - i; j++)
	 *(v+j) += x * *(yt+i+j);
      }
      else {
	*(v+i) += x * x;
      }
    }

    nframes++;
    t++;
    if (t == en)
      break;
  }

  p->nframes = nframes;
  cluster_const_set(p);

  return(p);
}

/* ---------------------------------------------------------------------------------- */
/* ----- cluster_t *build_clusters(cluster_t *, unsigned long *, unsigned long) ----- */
/* ---------------------------------------------------------------------------------- */
/*
 * Build clusters from initial clustering.
 */
cluster_t *build_clusters(cluster_t *clusters, unsigned long *nclusters, unsigned long nframes)
{
  cluster_t *p, *c[2];
  double dmin, d;
  icdistmat_t *icdmat;
  unsigned long step = 1;

  void print_cluster_name(FILE *, char *, cluster_t *, char *);

  if ((icdmat = distance_matrix_alloc(*nclusters)) == NULL) {
    fprintf(stderr, "scluster error -- cannot allocate memory\n");
    cluster_list_free(clusters);
    return(NULL);
  }

  p = clusters;

  while (*nclusters > mclusters) {
    
    if (kl2)
      distance_matrix_fill(icdmat, c, &dmin, p, cluster_kl2_distance, lambda);
    else if (ratio)
      distance_matrix_fill(icdmat, c, &dmin, p, cluster_like_distance, lambda);
    else
      distance_matrix_fill(icdmat, c, &dmin, p, cluster_bic_distance, lambda);

    /* 
      if the smallest distance between two clusters is more than the
      minimum inter-cluster distance (whic is also the maximum
      intra-cluster distance), don't force a merge that would do more
      harm than good!
    */

    if (trace > 2) {
      fprintf(stdout, "step %lu:\n", step);  
      fprintf(stdout, "  dmin=%-10.4f", dmin);
    }

    if (dmin > maxd)
      break;

    d = cluster_delta_bic(c[0], c[1], lambda, nframes);

    if (trace > 2)
      fprintf(stdout, "  dbic=%-10.4f", d);

    /*
      in the iterative version of the algorithm, search for the next
      best pair of clusters as long as delta BIC is positive.
    */
    if (iterative) {
 
      while (d > 0) {
	
	if (! distance_matrix_search(icdmat, c, &dmin))
	  break;

	if (trace > 2)
	  fprintf(stdout, "\n  dmin=%-10.4f", dmin);
	
	if (dmin > maxd)
	  break;
	
	d = cluster_delta_bic(c[0], c[1], lambda, nframes);

	if (trace > 2)
	  fprintf(stdout, "  dbic=%-10.4f", d);
      }
    }

    if (trace > 2)
      fprintf(stdout, "\n");

    /*
      if delta BIC is positive, even after the search of the best pair
      in iterative mode, then it is the end of the road.
    */
    if (d > 0)
      break;

    if (trace > 3) {
      print_cluster_name(stdout, "  ==> merge (", c[0], ")");
      print_cluster_name(stdout, " and (", c[1], ")\n");
    }

    p = cluster_merge(p, c[0], c[1]);
    *nclusters -= 1;
    step += 1;
  }

  return(p);
}

/* ------------------------------------------------------------------------ */
/* ----- void print_cluster_name(FILE *, char *, cluster_t *, char *) ----- */
/* ------------------------------------------------------------------------ */
void print_cluster_name(FILE *f, char *head, cluster_t *c, char *trail)
{
  seglink_t *p;
  asseg_t *seg;
  unsigned short i;
  int j = 0;

  fprintf(f, "%s", head);

  p = c->link;

  while (p) {
    seg = p->seg;

    if (j)
      fprintf(f, "/");
    else
      j = 1;

    for (i = 0; i < seg->label->nlabels; i++)
      fprintf(f, i ? "+%s" : "%s", seg->label->name[i]);

    p = p->next;
  }

  fprintf(f, "%s", trail);
}

/* ----------------------------------------------------------- */
/* ----- int make_segmentations(script_t *, cluster_t *) ----- */
/* ----------------------------------------------------------- */
/*
 * Make segmentations from clustering.
 */
int make_segmentations(script_t *script, cluster_t *c)
{
  unsigned long i, id;
  cluster_t *p = c;
  asseg_t *seg, *next;
  char name[10];
  seglink_t *link;

  /* change names */
  id = 1;
  while (p) {
    sprintf(name, "%s%lu", CLUSTER_NAME_STRING, id);
    
    link = p->link;
    while (link) {
      seg = link->seg;

      if (set_seg_label(seg, name) != 1) {
	fprintf(stderr, "scluster error -- cannot allocate memory\n");
	return(1);
      }

      link = link->next;
    }

    p = p->next;
    id += 1;
  }

  if (merge)   /* agglomerate adjacent segments with same label */
    for (i = 0; i < script->nfiles; i++) {
      seg = script->seg[i];
      
      while (seg) {
	
	next = get_seg_next(seg);
	
	if (next) {
	  if (seg_label_compare(get_seg_label(seg), get_seg_label(next)) == 0) {
	    set_seg_end_time(seg, get_seg_end_time(next));
	    set_seg_score(seg, get_seg_score(seg) + get_seg_score(next)); /* assume additive scores */
	    set_seg_next(seg, get_seg_next(next));
	    if (get_seg_next(next))
	      set_seg_prev(get_seg_next(next), seg);
	    seg_free(next);
	  }
	  else
	    seg = next;
	}
	else
	  seg = next;
      }
    }

  return(0);
}

/* ---------------------------------------------------------------------------------- */
/* ----- unsigned long write_script_seg(script_t *, const char *, const char *) ----- */
/* ---------------------------------------------------------------------------------- */
/*
 * Write script segmentations to disk.
 */
unsigned long write_script_seg(script_t *script, const char *outfn, const char *suffix)
{
  unsigned long i;
  char fn[MAX_FILE_NAME_LENGTH];
  const char *cp;

  for (i = 0; i < script->nfiles; i++) {

    if (! outfn) {
      strcpy(fn, script->segfn[i]);
      if (suffix) {
	strcat(fn, ".");
	strcat(fn, suffix);
      }
      cp = fn;
    }
    else 
      cp = outfn;
    
    if (trace) {
      fprintf(stdout, "  %s -> %s\n", script->segfn[i], cp);
      fflush(stdout);
    }

    if (seg_write(script->seg[i], cp, "lsep") == 0) {
      fprintf(stderr, "scluster error -- cannot write segmentation %lu to file %s\n", i+1, cp);
      return(i);
    }
  }

  return(i);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -k, --kullback               use symmetric KL distance
 *  -r, --ratio                  use likelihood ratio distance
 *  -l, --lambda=f               complexity penalty weight
 *  -f, --full                   use full covariance matrices
 *  -t, --threshold=f            minimum inter-cluster distance
 *  -d, --max-distance=f         distance above which clusters are not merged
 *  -n, --num-clusters=n         minimum number of clusters
 *  -i, --iterate                iterative mode
 *  -L, --file-list=fn           input script (none)
 *  -C, --convert=s              convertion string - ADNZ (none)
 *  -S, --input-bufsize=n        input buffer size in kbytes
 *  -x, --label=s                score segments with specified name (none)
 *  -b, --suffix=fn              output suffix (in place)
 *  -m, --merge                  merge adjacent segments in same cluster
 *  -v, --verbose                verbose mode
 *  -V, --version                print version number
 *  -h, --help                   help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"kullback", no_argument, NULL, 'k'},
    {"ratio", no_argument, NULL, 'r'},
    {"lambda", required_argument, NULL, 'l'},
#if HAVE_GSL
    {"full", no_argument, NULL, 'f'},
#endif
    {"max-distance", required_argument, NULL, 'd'},
    {"num-clusters", required_argument, NULL, 'n'},
    {"iterative", no_argument, NULL, 'i'},
    {"file-list", required_argument, NULL, 'L'},
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"label", required_argument, NULL, 'x'},
    {"suffix", required_argument, NULL, 'b'},
    {"merge", no_argument, NULL, 'm'},
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

#if HAVE_GSL
  while ((c = getopt_long(argc, argv, "krl:fd:n:iL:C:S:x:b:mv::Vh", longopts, &i)) != EOF)
#else
  while ((c = getopt_long(argc, argv, "krl:d:n:iL:C:S:x:b:mv::Vh", longopts, &i)) != EOF)
#endif
    switch (c) {
      
    case 'k':
      kl2 = 1;
      break;

    case 'r':
      ratio = 1;
      break;

    case 'l': 
      lambda = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "scluster error -- invalid argument %s to --lambda (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
#if HAVE_GSL
    case 'f':
      fullc = 1;
      break;
#endif

    case 'd':
      maxd = strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "scluster error -- invalid argument %s to --max-distance (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'n':
      mclusters = strtoul(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "scluster error -- invalid argument %s to --num-clusters (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'i':
      iterative = 1;
      break;
      
    case 'L':
      scpfn = optarg;
      break;

    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;
      
    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "scluster error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'x':
      labname = optarg;
      break;
      
    case 'b':
      suffix = optarg;
      break;

    case 'm':
      merge = 1;
      break;

    case 'v':
      trace = 1;
      if (optarg) {
        trace = (int)strtol(optarg, &p, 0);
        if (p == optarg || *p != 0) {
          fprintf(stderr, "scluster error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
          return(1);
        }
      }
      break;
      
    case 'V':
      version = 1;
      break;
      
    case 'h':
      help = 1;
      return(0);
      break;
      
    default:
      fprintf(stderr, "scluster error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  /* fool proofing */
  if (kl2 && ratio) {
    fprintf(stderr, ">>>>> WARNING >>>>> overrinding option --ratio with --kullback in scluster\n");
    ratio = 0;
  }

  return(0);
}

#undef _scluster_c_
