/******************************************************************************/
/*                                                                            */
/*                                sgmestim.c                                  */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/

/*
 * Gaussian mixture model parameter estimation.
 *
 * This is the main parameter estimation program which implements both
 * maximum likelihood (ML) and maximum a posteriori (MAP) criteria.
 *
 * The program takes as input an initial model and either a collection
 * of datafiles or, alternately a training script. The training script
 * is a text file containing a list of data files to use for paramter
 * estimation, each line following the syntax
 *
 *   /path/to/spro/data/file [path/to/the/corresponding/label/file]
 *
 * where the label file is optionnal.
 *
 * The estimation algorithms iterates over the specified training data
 * files, until the maximum number of iteration is reached or until
 * the relative increase of the average frame log-likelihood over the
 * data is below the epsilon threshold if the latter is specified.
 *
 * If a label is specified with the --label=s option, only the
 * segments corresponding to that label are used for
 * re-estimation. This imply that a training script is specified and
 * that the segmentation files have been specified in the training
 * script.
 *
 * Last but not least, the --update option makes it possible to
 * re-estimate the model parameters partially. For example,
 * --update=wm would only re-estimate the weights and the mean vetors,
 * leaving the covariance vectors unchanged.
 *
 * Trace levels:
 *  1    basic reporting
 *  2    basic EM iterations
 *  3    detailed EM accumulation
 */

#define _sgmestim_c_

#include <audioseg.h>
#include <script.h>
#include <getopt.h>

static char *cvsid = "$Id: sgmestim.c,v 1.11 2003/11/20 15:50:58 ggravier Exp $";

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
char *labname = NULL;          /* extract label name                   */
char *scpfn = NULL;            /* training script file name            */
int maxiter = 10;              /* maximum number of iterations         */
double epsilon = 0.0;          /* main loop stopper                    */
double vfloor = 0.0001;        /* variance floor                       */
double wfloor = 0.0;           /* weight floor                         */
double map = 0.0;              /* MAP parameter                        */
char *ofn = NULL;              /* output filename                      */
char *varfn = NULL;            /* variance filename                    */
int update = W_UP | M_UP | V_UP; /* what do we wanna update?           */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int trace = 0;                 /* trace level                          */
int help = 0;                  /* print usage                          */
int version = 0;               /* version flag                         */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sgmestim [options] model data ...\n");
  fprintf(stdout, "  sgmestim [options] --file-list=script model\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Gaussian mixture model parameter estimation using the ML or MAP\n");
  fprintf(stdout, "  criteria. The input script file contains a list of all the training\n");
  fprintf(stdout, "  data files, optionally with the corresponding segmentation files\n");
  fprintf(stdout, "  (required if --label is specified).\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -m, --map=f                  MAP prior weight (%.2e)\n", map);
  fprintf(stdout, "  -u, --update=s               model update string (wmv)\n");
  fprintf(stdout, "  -i, --max-iter=n             maximum number of iterations (%d)\n", maxiter);
  fprintf(stdout, "  -e, --epsilon=f              convergence factor (%.2e)\n", epsilon);
  fprintf(stdout, "\n");
  fprintf(stdout, "  -w, --weight-floor=f         weight floor (%.2e)\n", wfloor);
  fprintf(stdout, "  -f, --variance-floor=f       variance floor (%.2e)\n", vfloor);
  fprintf(stdout, "  -c, --variance=fn            covariance file (nil)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -L, --file-list=fn           input script (none)\n");
  fprintf(stdout, "  -x, --label=s                label (none)\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ADNZ (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);
  fprintf(stdout, "\n");
  fprintf(stdout, "  -o, --output=fn              output model filename (same as input)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char** argv)
{
  gmm_t *gmm;                       /* model                                 */
  script_t *script;                  /* training script                       */
  seglab_t *label = NULL;            /* label to chop                         */
  char *modfn;                       /* initial model file name               */

  int read_args(int, char **);
  int estimate(gmm_t *, script_t *, seglab_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }

  /* reading arguments. */
  if (optind < argc)
    modfn = argv[optind++];
  else {
    fprintf(stderr, "sgmestim error -- no model filename specified (use --help to get usage)\n");
    return(1);
  }

  if (! scpfn) {
    if (optind < argc) {
      if ((script = make_script(argc - optind, argv + optind)) == NULL) {
	fprintf(stderr, "sgmestim error -- cannot create script from arguments\n");
	return(1);
      }
    }
    else {
      fprintf(stderr, "sgmestim error -- no data filename specified (use --help to get usage)\n");
      return(1);
    }
  }
  else if ((script = load_script(scpfn, 0)) == NULL) {
    fprintf(stderr, "sgmestim error -- cannot load script from file %s\n", scpfn);
    return(1);
  }
  
  /* ----- load segmentations if necessary ----- */
  if (labname) {
    if ((label = str_to_label(labname)) == NULL) {
      fprintf(stderr, "sgmestim error -- cannot create reference label\n");
      free_script(script);
      return(1);
    }

    if (load_script_seg(script) == NULL) {
      fprintf(stderr, "sgmestim error -- cannot load segmentation\n");
      free_script(script); seg_label_free(label);
      return(1);
    }
  }

  /* ----- load the input model ----- */
  if ((gmm = gmm_read(modfn)) == NULL) {
    fprintf(stderr, "sgmestim error -- cannot read model from file %s\n", modfn);
    free_script(script); seg_label_free(label);
    return (1);
  }

  /* ----- run the estimation ----- */
  if (trace) {
    fprintf(stdout, "script    = %s (%lu files)\n", scpfn ? scpfn : "<nil>", script->nfiles);
    fprintf(stdout, "label     = %s\n", labname ? labname : "<nil>");
    fprintf(stdout, "criterion = %s\n", map > 0.0 ? "MAP" : "ML");
    fflush(stdout);
  }

  if (estimate(gmm, script, label) != 0) {
    free_script(script); gmm_free(gmm); seg_label_free(label);
    return(1);
  }

  /* ----- write model to output file ----- */
  ofn = ofn ? ofn : modfn;

  if (trace) {
    fprintf(stdout, "saving model to file %s\n", ofn);
    fflush(stdout);
  }

  if (gmm_write(ofn, gmm) != 0) {
    fprintf(stderr, "sgmestim error -- cannot write model to output file %s", ofn);
    free_script(script); gmm_free(gmm); seg_label_free(label);
    return(1);
  }

  /* ----- clean memory and go back to bed ----- */
  seg_label_free(label);
  free_script(script); 
  gmm_free(gmm);

  return(0);
}

/* --------------------------------------------------------- */
/* ----- int estimate(gmm_t *, script_t *, seglab_t *) ----- */
/* --------------------------------------------------------- */
int estimate(gmm_t *gmm, script_t *script, seglab_t *label)
{
  spfstream_t *f;                    /* input feature stream                  */
  gmm_t *prior = NULL;               /* MAP prior model                       */
  gmmacc_t *accs;                    /* accumulators                          */
  double seglogp;                    /* segment log-likelihood                */
  double llk, llkold = 0.0;          /* total data log-likelihood             */
  unsigned long nframes, n;          /* number of frames                      */
  double *varfloor = NULL;           /* variance floor vector                 */
  asseg_t *curseg;
  float st, et;
  int i, it;
  unsigned long sn, en;
  
  gmm_t *duplicate(gmm_t *);
  asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int);

  /* ----- initialization ----- */
  if ((accs = gmm_acc_alloc(get_gmm_num_comp(gmm), get_gmm_dim(gmm))) == NULL) {
    fprintf(stderr, "sgmestim error -- cannot allocate memory for the accumulators\n");
    return(1);
  }

  if (map > 0.0) {
    /* copy initial model to keep track of the prior parameters */
    if ((prior = duplicate(gmm)) == NULL) {
      fprintf(stderr , "sgmestim error -- cannot duplicate initial model\n");
      gmm_acc_free(accs);
      return(1);
    }
  }

  /* ----- variance floor initialization ----- */
  if (vfloor > 0.0) {
    if ((varfloor = set_var_floor(vfloor, varfn, get_gmm_dim(gmm))) == NULL) {
      gmm_acc_free(accs); gmm_free(prior);
      return(1);      
    }
  }

  /* ----- iterate ----- */
  for (it = 0; it < maxiter; it++) {

    llk = 0.0;
    nframes = 0;
    gmm_acc_reset(accs);
    
    /* accumulate stats for over all the training files */
    for (i = 0; i < script->nfiles; i++) {

      curseg = get_next_seg(&st, &et, label ? script->seg[i] : NULL, label, 1);

      if (st == ASEG_NULL_TIME)
	continue;

      /* open feature stream */
      if ((f = spf_input_stream_open(script->datfn[i], cvtflag, iobs)) == NULL) {
	fprintf(stderr, "sgmestim error -- cannot open feature stream %s\n", script->datfn[i]);
	return(1);
      }

      /* check it's a valid file */
      if (spf_stream_dim(f) != get_gmm_dim(gmm)) {
	fprintf(stderr, 
		">>>>> WARNING >>>>> incompatible data (%d) and model (%d) dimensions in file %s (ignoring)\n", 
		spf_stream_dim(f), get_gmm_dim(gmm), script->datfn[i]);
	continue;
      }
      
      /* loop on all segments */
      while (st != ASEG_NULL_TIME) {
	
	/* get initial start and possibly end frame number and seek in
	   input stream */
	sn = (unsigned long)(st * spf_stream_rate(f));
	if (et != ASEG_NULL_TIME) {
	  en = (unsigned long)(et * spf_stream_rate(f));
	  n = (en - sn > 0) ? (en - sn) : 1;
	}
	else
	  n = 0;

	if (spf_stream_seek(f, sn, SEEK_SET) != 0) {
	  fprintf(stderr, "sgmestim error -- cannot go to frame %lu at t=%.f in file %s\n", sn, st, script->datfn[i]);
	  spf_stream_close(f);
	  return(1);
	}
	
	seglogp = gmm_accumulate(f, &n, gmm, accs);
	llk += seglogp;
	nframes += n;
	
	if (trace > 2) {
	  fprintf(stdout, "   file=%s  st=%.2f  et=%.2f nframes=%lu  logp=%e\n", script->datfn[i], st, et, n, seglogp / n);
	  fflush(stdout);
	}

	/* switch to next segment */
	curseg = get_next_seg(&st, &et, curseg, label, 0);      
      }

      spf_stream_close(f);
    }

    /* if there is no data for the model or all the data lead to out
       of bound values during the frame log like computation */
    if (nframes == 0) {
      fprintf(stderr, "sgmestim error -- no training data available\n");
      return(1);
    }

    if ((it > 1) && (fabs((llk - llkold) / llkold) < epsilon))
      break;
    
    if (trace > 1) {
      fprintf(stdout, "Iteration %d: nframes=%lu  logp=%e", it, nframes, llk / (double)nframes);
      if (it)
	fprintf(stdout, "  diff=%e  increase=%e", (llk - llkold) / (double)nframes, fabs((llk - llkold) / llkold));
      fprintf(stdout, "\n");
      fflush(stdout);
    }
      
    if (map > 0.0)
      gmm_map_update(gmm, accs, nframes, update, wfloor, varfloor, map, prior);
    else
      gmm_ml_update(gmm, accs, nframes, update, wfloor, varfloor);

    /* check defunct components */
    n = 0;
    for (i = 0; i < get_gmm_num_comp(gmm); i++)
      if (gmm->w[i] < GMM_MIN_WEIGHT)
	n++;
    if (n)
      fprintf(stderr, ">>>>> WARNING >>>>> %lu defunct component after iteration %d\n", n, it);

    llkold = llk;
  }

  /* ----- clean memory ----- */
  gmm_free(prior);
  gmm_acc_free(accs);
  if (varfloor)
    free(varfloor);

  return(0);
}

/* ------------------------------------------------------------------------------- */
/* ----- asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int) ----- */
/* ------------------------------------------------------------------------------- */
/*
 * Get next segment start and end times. If no more segments, set st
 * to ASEG_NULL_TIME. If end of segment is unkown (i.e. it matches the
 * end of the input feature stream) et is set to ASEG_NULL_TIME. In
 * any other cases, st and et are set to the segment start and end
 * times (in seconds). If a segmentation is supplied, move to the next
 * segment in the segmentation, searching for the specified label if
 * there's one. In this case, the function returns a pointer to the
 * segment found to be used in successive calls to get_next_seg().
 *
 * The function is initialized when first invoked with a non null
 * reset flag.
 */
asseg_t *get_next_seg(float *st, float *et, asseg_t *seg, seglab_t *label, int reset)
{
  asseg_t *p = NULL;


  if (seg) {
    p = (reset) ? seg : seg->next;
    if (! p) {
      *st = *et = ASEG_NULL_TIME;
      return(NULL);
    }
    while (seg_label_compare(p->label, label) != 0)
      if ((p = p->next) == NULL) {
	*st = *et = ASEG_NULL_TIME;
	return(NULL);
      }
    *st = p->st;
    *et = p->et;
  }
  else {
    *st = (reset) ? 0.0 : ASEG_NULL_TIME;
    *et = ASEG_NULL_TIME;
  }

  return(p);
}

/* --------------------------------------- */
/* ----- gmm_t *duplicate(gmm_t *) ----- */
/* --------------------------------------- */
/*
 * Duplicate a model.
 */
gmm_t *duplicate(gmm_t *p)
{
  gmm_t *newp;
  unsigned short i;
  unsigned short ncomp, dim;

  ncomp = get_gmm_num_comp(p);
  dim = get_gmm_dim(p);

  if ((newp = gmm_alloc(ncomp, dim)) == NULL)
    return(NULL);

  for (i = 0; i < ncomp; i++) {
    set_gmm_comp_weight(newp, i, get_gmm_comp_weight(p, i));
    set_gmm_comp_logc(newp, i, get_gmm_comp_logc(p, i));
    set_gmm_comp_expz(newp, i, get_gmm_comp_expz(p, i));
    memcpy(get_gmm_comp_mean(newp, i), get_gmm_comp_mean(p, i), dim * SIZEOF_FLOAT);
    memcpy(get_gmm_comp_ivar(newp, i), get_gmm_comp_ivar(p, i), dim * SIZEOF_FLOAT);
  }

  return(newp);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -L, --file-list=fn           input script
 *  -l, --label=s                label
 *  -n, --max-iter=n             maximum number of iterations
 *  -e, --epsilon=f              convergence factor
 *  -w, --weight-floor=f         weight floor
 *  -f, --variance-floor=f       variance floor
 *  -c, --variance=fn            covariance vector
 *  -u, --update=s               model update string
 *  -o, --output=fn              output filename
 *  -C, --convert=s              convertion string
 *  -S, --input-bufsize=n        input buffer size in kbytes
 *  -m, --map=f                  MAP posterior weight
 *  -v, --verbose                verbose mode
 *  -V, --version                print version number
 *  -h, --help                   this help message
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"file-list", required_argument, NULL, 'L'},
    {"output", required_argument, NULL, 'o'},
    {"label", required_argument, NULL, 'x'},
    {"weight-floor", required_argument, NULL, 'w'},
    {"var-floor", required_argument, NULL, 'f'},
    {"variance", required_argument, NULL, 'c'},
    {"max-iter", required_argument, NULL, 'i'},
    {"epsilon", required_argument, NULL, 'e'},
    {"update", required_argument, NULL, 'u'},
    {"map", required_argument, NULL, 'm'},
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "L:x:i:e:c:w:f:u:o:C:S:m:v::Vh", longopts, &i)) != EOF)
    switch (c) {

    case 'x':
      labname = optarg;
      break;

    case 'L':
      scpfn = optarg;
      break;

    case 'i':
      maxiter = (int)strtol(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --max-iter (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'e':
      epsilon = (double)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --epsilon (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'f':
      vfloor = (double)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --varfloor (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'c':
      varfn = optarg;
      break;

    case 'w':
      wfloor = (double)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --weightfloor (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'u':
      p = optarg;
      update = 0;
      while (*p) {
	switch (*p) {
	case 'w': 
	case 'W':
	  update |= W_UP; 
	  break;

	case 'm': 
	case 'M':
	  update |= M_UP;
	  break;

	case 'v': 
	case 'V':
	  update |= V_UP;
	  break;

	default:
	  fprintf(stderr, "sgmestim error -- invalid character %c in update string (use --help to get usage)\n", *p);
	  return(1);
	}
	p++;
      }
      break;

    case 'o':
      ofn = optarg;
      break;

    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;

    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'm':
      map = (double)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmestim error -- invalid argument %s to --map (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'v':
      trace = 1;
      if (optarg) {
	trace = strtol(optarg, &p, 0);
	if (p == optarg || *p != 0) {
	  fprintf(stderr, "sgmestim error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      break;

    case 'V':
      version = 1;
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sgmestim error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  /* basic option sanity check */
  if (vfloor < 0.0) {
    fprintf(stderr, "sgmestim error -- invalid variance floor value %e (not positive)\n", vfloor);
    return(1);
  }

  if (wfloor < 0.0) {
    fprintf(stderr, "sgmestim error -- invalid weight floor value %e (not positive)\n", vfloor);
    return(1);
  }

  if (map < 0.0) {
    fprintf(stderr, "sgmestim error -- invalid MAP prior weight value %.2f (not positive)\n", map);
    return(1);
  }

  return(0);
}

#undef _sgmestim_c_
