/******************************************************************************/
/*                                                                            */
/*                                  ssad.c                                    */
/*                                                                            */
/*                        Audio Segmentation Toolkit                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.9 $
 *
 */

/*
 * Signal activity detector.
 *
 * Signal activity detector based on a bi-gaussian model of the frame
 * energy distribution and a maximum likelihood criterion for
 * classification. The program outputs a segmentation wich identifies
 * either the silence segments, or the audio signal segments or even
 * both. A minimum silence segment length can be specified to avoid
 * over-segmentation problems due to the processing at the frame
 * level.
 */

#define _ssad_c_

#include <audioseg.h>
#include <getopt.h>

static char *cvsid = "$Id: ssad.c,v 1.9 2003/11/20 15:50:58 ggravier Exp $";

/* output strings */
#define SILENCE_STRING "sil"
#define SPEECH_STRING  "speech"

/* output modes */
#define UNKNOWN 0
#define SILENCE 1
#define SPEECH 2
#define BOTH 3

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
int format = SPRO_SIG_PCM16_FORMAT; /* signal file format                     */
float Fs = 8000.0;                /* input signal sample rate                 */
int channel = 1;                  /* channel to process                       */
int swap = 0;                     /* change input sample byte order           */
size_t ibs = 10000000;            /* input buffer size                        */
float fm_l = 20.0;                /* frame length in ms                       */
float fm_d = 10.0;                /* frame shift in ms                        */
int win = 0;                      /* weighting window                         */
float threshold = 0.0;            /* deviation w.r.t. standard deviation      */
float minlen = 0.0;               /* minimum silence segment length           */
int ofmt = SILENCE;               /* output format                            */
int uselog = 0;                   /* use log-energy rather than energy        */
int trace = 0;                    /* trace level                              */
int help = 0;                     /* print usage                              */
int version = 0;                  /* print version information                */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "    ssad [options] ifile ofile\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "    Silence detector.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -F, --format=s               input signal file format (pcm16)\n");
  fprintf(stdout, "  -f, --sample-rate=n          signal sample rate for pcm16 input (%.1f Hz)\n", Fs);
  fprintf(stdout, "  -x, --channel=n              channel number (%d)\n", channel);
  fprintf(stdout, "  -B, --swap                   swap sample byte order (don't)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (10000)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -s, --log                    use log-energy (energy)\n");
  fprintf(stdout, "  -l, --length=n               frame length in ms (%.1f ms)\n", fm_l);
  fprintf(stdout, "  -d, --shift=n                frame shift in ms (%.1f ms)\n", fm_d);
  fprintf(stdout, "  -w, --window=s               weighting window (none)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -t, --threshold=f            set signal energy threshold to m1 - f * s1 (ML)\n");
  fprintf(stdout, "  -m, --min-length=f           minimum silence segment length in seconds (%.1f)\n", minlen);
  fprintf(stdout, "  -r, --reverse                output speech segments (silence)\n");
  fprintf(stdout, "  -a, --all                    output speech and silence segments\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* ---------------------------------- */
/* ----- local type definitions ----- */
/* ---------------------------------- */
typedef struct {
  float m[2];
  float v[2];
  double c[2];
} bigauss_t;

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char **argv)
{
  char *ifn, *ofn;                /* I/O filenames                            */
  sigstream_t *s;                 /* input signal stream                      */
  asseg_t *seg;

  int read_args(int, char **);
  asseg_t *silence_detection(sigstream_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }

  if (optind < argc)
    ifn = argv[optind++];
  else {
    fprintf(stderr, "ssad error -- no input filename specified (use --help to get usage)\n");
    return(1);
  }

  if (optind < argc)
    ofn = argv[optind++];
  else {
    fprintf(stderr, "ssad error -- no output filename specified (use --help to get usage)\n");
    return(1);
  }

  if (optind < argc) {
    fprintf(stderr, "ssad error -- invalid number of arguments (use --help to get usage)\n");
    return(1);
  }
   

  /* ---- open input stream ----- */
  if (trace) {
    fprintf(stdout, "%s --> %s\n", (ifn) ? (ifn) : "stdin", (ofn) ? (ofn) : "stdout");
    fflush(stdout);
  }
  
  if ((s = sig_stream_open(ifn, format, Fs, ibs, swap)) == NULL) {
    fprintf(stderr, "ssad error -- cannot open input signal stream %s\n", (ifn) ? (ifn) : "stdin");
    return(1);
  }
  
  /* ----- run silence detector ----- */
  if (trace > 1) {
    fprintf(stdout, "silence detection ... \n");
    fflush(stdout);
  }
  if ((seg = silence_detection(s)) == NULL)
    return(1);

  /* ----- output segmentation to file ----- */
  if (trace > 1) {
    fprintf(stdout, "writing segmentation to output file ... \n");
    fflush(stdout);
  }
  if (seg_write(seg, ofn, "LSE") == 0) {
    fprintf(stderr, "ssad error -- cannot write output segmentation to %s\n", (ofn) ? (ofn) : "stdout");
      sig_stream_close(s); seg_list_free(seg);
    return(1);
  }

  /* ----- clean memory ----- */
  sig_stream_close(s);
  seg_list_free(seg);
  
  return(0);
}

/* ----------------------------------------------------- */
/* ----- asseg_t *silence_detection(sigstream_t *) ----- */
/* ----------------------------------------------------- */
/*
 * process input file.
 */
asseg_t *silence_detection(sigstream_t *s)
{
  bigauss_t bg;
  spfbuf_t *e;
  asseg_t *seg;
  unsigned short nl, nd;
  double emin, emax;

  spfbuf_t *get_energy_profile(sigstream_t *, unsigned short, unsigned short, double *, double *);
  void init_bigauss(bigauss_t *, double, double);
  void buf_to_bigauss(spfbuf_t *, bigauss_t *, int, double);
  asseg_t *profile_to_seg(spfbuf_t *, bigauss_t *, float);

  
  /* ----- compute energy profile ----- */
  if (trace > 1) {
    fprintf(stdout, "  computing energy profile ... \n");
    fflush(stdout);
  }

  nl = (unsigned short)(fm_l * s->Fs / 1000.0);
  nd = (unsigned short)(fm_d * s->Fs / 1000.0);

  if ((e = get_energy_profile(s, nl, nd, &emin, &emax)) == NULL)
    return(NULL);

  /* ----- map gaussians to profile ----- */
  if (trace > 1) {
    fprintf(stdout, "  gaussian mapping ... \n");
    fflush(stdout);
  }
  init_bigauss(&bg, emin, emax);
  buf_to_bigauss(e, &bg, 20, 0.0001);

  /* ----- convert profile to segmentation ----- */
  if ((seg = profile_to_seg(e, &bg, nd / s->Fs)) == NULL) {
    fprintf(stderr, "ssad error -- cannot create output segmentation\n");
    spf_buf_free(e); seg_list_free(seg);
    return(NULL);
  }

  spf_buf_free(e);

  return(seg);
}

/* ---------------------------------------------------------------------------- */
/* ----- spfbuf_t *get_energy_profile(sigstream_t *, unsigned short,      ----- */
/* -----                              unsigned short, double *, double *) ----- */
/* ---------------------------------------------------------------------------- */
/*
 * Compute signal stream energy profile.
 */
spfbuf_t *get_energy_profile(sigstream_t *s, unsigned short l, unsigned short d, double *emin, double *emax)
{
  spfbuf_t *buf;
  unsigned long n;
  float *w = NULL;
  sample_t *sbuf;
  spsig_t *frame;
  spf_t e;

  /* ----- initialize some more stuff ----- */
  *emax = FLT_MIN;
  *emin = FLT_MAX;
  
  if ((frame = sig_alloc(l)) == NULL) {
    fprintf(stderr, "ssad error -- cannot allocate frame signal buffer\n");
    return(NULL);
  }

  if ((buf = spf_buf_alloc(1, 40000)) == NULL) {
    fprintf(stderr, "ssad error -- cannot allocate output feature buffer\n");
    sig_free(frame);
    return(NULL);
  }

  if (win) {
    if ((sbuf = (sample_t *)malloc(l * sizeof(sample_t))) == NULL) {
      fprintf(stderr, "ssad error -- cannot allocate memory\n");
      sig_free(frame); spf_buf_free(buf);
      return(NULL);
    }
    if ((w = set_sig_win(l, win)) == NULL) {
      fprintf(stderr, "ssad error -- cannot allocate weighting window\n");
      free(sbuf); sig_free(frame); spf_buf_free(buf);
      return(NULL);    
    }
  }
  else
    sbuf = frame->s;

  /* ----- compute profile ----- */
  n = 0;
  while (get_next_sig_frame(s, channel, l, d, 0.0, sbuf)) {

    /* weight signal */
    if (w)
      sig_weight(frame, sbuf, w);

    /* compute frame energy */
    e = (spf_t)sig_normalize(frame, 0);

    if (uselog) 
      e = (e < SPRO_ENERGY_FLOOR) ? (spf_t)log(SPRO_ENERGY_FLOOR) : (spf_t)log(e);

    if (spf_buf_append(buf, &e, 1, 10000) == NULL) {
      fprintf(stderr, "ssad error -- cannot append energy value to output feature buffer\n");
      free(sbuf); sig_free(frame); if (win) {spf_buf_free(buf); free(w);}
      return(NULL);
    }
    
    if (e > *emax)
      *emax = e;
    if (e < *emin)
      *emin = e;
  }

  /* ---- clean and get out of here! ----- */
  sig_free(frame); 
  if (win) {
    free(sbuf); 
    free(w);
  }

  return(buf);
}

/* ---------------------------------------------------------- */
/* ----- void init_bigauss(bigauss_t *, double, double) ----- */
/* ---------------------------------------------------------- */
/*
 * Initialize bi Gaussian model.
 */
void init_bigauss(bigauss_t *bg, double emin, double emax)
{
  double d = (emax - emin) / 8.0; 

  bg->m[0] = emin + d;
  bg->v[0] = 1.0; 
  
  bg->m[1] = emax - d;
  bg->v[1] = 1.0; /* v = 1/v */

  bg->c[0] = -0.5 * bg->m[0] * bg->m[0]; /* c = 0.5 * ( ln(1/v) - 1/v * m * m ) */ 
  bg->c[1] = -0.5 * bg->m[1] * bg->m[1];
}

/* ---------------------------------------------------------------------- */
/* -----  void buf_to_bigauss(spfbuf_t *, bigauss_t *, int, double) ----- */
/* ---------------------------------------------------------------------- */
/*
 * Map buffer features to bi-gaussian.
 */
void buf_to_bigauss(spfbuf_t *e, bigauss_t *bg, int maxiter, double epsilon)
{
  unsigned long i, t, n1, n2;
  double m1, v1, m2, v2; /* accumulators */
  double m;
  double llk1, llk2; /* log-likelihoods */
  double v, vv; /* sample value */
  double llk, llkmem = 0.0; /* total data log-probability */

  i = 0;

  while (i < maxiter) {
    /* set accumulators to zero */
    n1 = n2 = 0;
    m1 = m2 = v1 = v2 = 0.0;
    llk = 0;

    /* loop on energy profile samples */
    for (t = 0; t < e->n; t++) {
      v = *((e->s)+t);
      vv = v * v;

      llk1 = bg->v[0] * bg->m[0] * v - 0.5 * bg->v[0] * vv + bg->c[0];
      llk2 = bg->v[1] * bg->m[1] * v - 0.5 * bg->v[1] * vv + bg->c[1];

      if (llk1 > llk2) { /* mapped to gaussian #1 */
	m1 += v;
	v1 += vv;
	n1++;
	llk += llk1;
      }
      else { /* mapped to gaussian #2 */
	m2 += v;
	v2 += vv;
	n2++;
	llk += llk2;
      }
    }

    /* trace some stuff */
    if (trace > 2) {
      fprintf(stdout, "   iteration %2lu: logp=%e  n1=%lu  n2=%lu\n", i, llk / (double)(e->n), n1, n2);
      fflush(stdout);
    }
    
    if (i)
      if ((llkmem - llk) / llkmem < epsilon)
	break;

    /* update models */
    m = m1 / (double)n1;
    v = (double)n1 / (v1 - m * m1); /* this is the inverse of the variance estimator */
    bg->m[0] = m;
    bg->v[0] = v;
    bg->c[0] = 0.5 * (log(v) - m * m * v) ;

    m = m2 / (double)n2;
    v = (double)n2 / (v2 - m * m2); /* this is the inverse of the variance estimator */
    bg->m[1] = m;
    bg->v[1] = v;
    bg->c[1] = 0.5 * (log(v) - m * m * v) ;

    i++;
  }

  if (trace > 2) {
    fprintf(stdout, "   silence: m=%-10.4f  v=%-10.4f\n", bg->m[0], 1.0 / bg->v[0]);
    fprintf(stdout, "   speech:  m=%-10.4f  v=%-10.4f\n", bg->m[1], 1.0 / bg->v[1]);
  }
}

/* ---------------------------------------------------------------------------- */
/* ------ asseg_t *profile_to_seg(spfbuf_t *, bigauss_t *, char *, float) ----- */
/* ---------------------------------------------------------------------------- */
/*
 * Create segmentation from features and the two gaussians.
 */
asseg_t *profile_to_seg(spfbuf_t *e, bigauss_t *bg, float frate)
{
  unsigned long i;
  float st1, et1; /* silence segment start and end times */
  float st2, et2; /* speech segment start and end times */
  unsigned long ns1, ns2;
  double d1, d2;
  asseg_t *seg = NULL;
  int state, label;
  double v, vv, logp1, logp2;
  
  asseg_t *add_seg(asseg_t **, float, float, int);
  
  state = UNKNOWN;
  st1 = st2 = 0.0;
  et1 = et2 = ASEG_NULL_TIME;
  ns1 = ns2 = 0;
  d1 = d2 = 0.0;

  for (i = 0; i < e->n; i++) {

    v = *(e->s+i);

    if (threshold != 0.0)
      label = (v < bg->m[1] - threshold *  sqrt(1.0 / bg->v[1])) ? (SILENCE) : (SPEECH);
    else {
      vv = v * v;    
      logp1 = bg->v[0] * bg->m[0] * v - 0.5 * bg->v[0] * vv + bg->c[0];
      logp2 = bg->v[1] * bg->m[1] * v - 0.5 * bg->v[1] * vv + bg->c[1];

      label = (logp1 > logp2) ? (SILENCE) : (SPEECH);
    }

    if (state == SPEECH && label == SILENCE) { /* potential end of a signal segment */
      et2 = (float)i * frate; /* detected a [st2,et2] speech segment */
      /* fprintf(stderr, " detected speech st=%.2f  et=%.2f\n", st2, et2); */
      st1 = et2; /* start new silence segment */
    }
    else if (state == SILENCE && label == SPEECH) { /* potential end of a silence segment */
      et1 = (float)i * frate; /* detected a [st1,et1] silence segment */
      /* fprintf(stderr, " detected silence st=%.2f  et=%.2f\n", st1, et1); */

      /* if segment is long enough, add the previous speech segment
	 and the current silence segment. Else, simply ignore the
	 silence segment and proceed... */

      if (et1 - st1 > minlen) {
	
	/* add previous speech segment if there was one */
	if (ofmt & SPEECH && et2 != ASEG_NULL_TIME) {
	  /* fprintf(stderr, " adding speech st=%.2f  et=%.2f\n", st2, et2); */
	  ns2++;
	  d2 += (et2 - st2);
	  if (add_seg(&seg, st2, et2, SPEECH) == NULL) {
	    seg_list_free(seg);
	    return(NULL);
	  }
	}
	
	/* add current silence segment */
	if (ofmt & SILENCE) {
	  ns1++;
	  d1 += (et1 - st1);
	  if (add_seg(&seg, st1, et1, SILENCE) == NULL) {
	    seg_list_free(seg);
	    return(NULL);
	  }
	}
	st2 = et1; /* set start of new speech segment */
      }
    
    }
    state = label;
  }

  /* check out if there's a last segment to output */
  if (state == SILENCE) {
      et1 = (float)i * frate; /* detected a [st1,et1] silence segment */
      
      /* if segment is long enough, add the previous speech segment
	 and the current silence segment. Else, add the last speech
	 segment. */
      
      if (et1 - st1 > minlen) {
	
	/* add previous speech segment */
	if (ofmt & SPEECH && et2 != ASEG_NULL_TIME) {

	  if (add_seg(&seg, st2, et2, SPEECH) == NULL) {
	    seg_list_free(seg);
	    return(NULL);
	  }
	  
	  ns2++;
	  d2 += (et2 - st2);
	}
	
	/* add current silence segment */
	if (ofmt & SILENCE) {

	  if (add_seg(&seg, st1, et1, SILENCE) == NULL) {
	    seg_list_free(seg);
	    return(NULL);
	  }

	  ns1++;
	  d1 += (et1 - st1);
	}
      }
      else if (ofmt & SPEECH) {
	et2 = et1;

	if (add_seg(&seg, st2, et2, SPEECH) == NULL) {
	  seg_list_free(seg);
	  return(NULL);
	}
	
	ns2++;
	d2 += (et2 - st2);
      }
  }
  else if (ofmt & SPEECH) {
    et2 = (float)i * frate; /* detected a [st2,et2] speech segment */

    if (add_seg(&seg, st2, et2, SPEECH) == NULL) {
      seg_list_free(seg);
      return(NULL);
    }    

    ns2++;
    d2 += (et2 - st2);
  }
  
  if (trace > 1) {
    if (ofmt & SILENCE) {
      fprintf(stdout, "number of silence segments = %ld\n", ns1);
      fprintf(stdout, "average silence duration = %f s\n", d1 / ns1);
    }
    if (ofmt & SPEECH) {
      fprintf(stdout, "number of speech segments = %ld\n", ns2);
      fprintf(stdout, "average speech duration = %f s\n", d2 / ns2);
    }
    fflush(stdout);
  }

  return(seg);
}

/* ----------------------------------------------------------- */
/* ----- asseg_t *add_seg(asseg_t **, float, float, int) ----- */
/* ----------------------------------------------------------- */
/*
 * Add a segment after the current segment, returning the adress of
 * the new segment.
 */
asseg_t *add_seg(asseg_t **seg, float st, float et, int label)
{
  static asseg_t *last = NULL;
  asseg_t *p;
  char silstr[] = SILENCE_STRING;
  char sigstr[] = SPEECH_STRING;

  if (label == SILENCE)
    p = seg_create(silstr, st, et, 0.0);
  else
    p = seg_create(sigstr, st, et, 0.0);

  if (*seg == NULL)
    *seg = p;
  
  if (last)
    last->next = p;
  
  last = p;
  
  return(p);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -F, --format=s            signal file format (pcm)
 *  -f, --sample-rate=f       signal sample rate (8000 Hz)
 *  -x, --channel=n           use channel number (n) (1)
 *  -b, --swap                swap input stream samples
 *  -M, --input-bufsize=n     input buffer size (in bytes)
 *  -l, --length=f            frame length in ms (20 ms)
 *  -d, --shift=f             frame shift in ms (10 ms)
 *  -w, --window=s            weighting window (none)
 *  -m, --min-length=n        minimum silence segment length
 *  -r, --reverse             output speech segments
 *  -a, --all                 output both speech and silence segments
 *  -v, --verbose[=n]         verbose mode
 *  -V, --version             print version number
 *  -h, --help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"format", required_argument, NULL, 'F'},
    {"sample-rate", required_argument, NULL, 'f'},
    {"channel", required_argument, NULL, 'x'},
    {"swap", no_argument, NULL, 'B'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"log", no_argument, NULL, 's'},
    {"length", required_argument, NULL, 'l'},
    {"shift", required_argument, NULL, 'd'},
    {"window", required_argument, NULL, 'w'},
    {"threshold", required_argument, NULL, 't'},
    {"min-length", required_argument, NULL, 'm'},
    {"reverse", no_argument, NULL, 'r'},
    {"all", no_argument, NULL, 'a'},
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "F:f:x:BS:sl:d:w:t:m:rav::Vh", longopts, &i)) != EOF)
    switch (c) {

    case 'F':
      if (strcasecmp(optarg, "pcm16") == 0)
	format = SPRO_SIG_PCM16_FORMAT;
      else if (strcasecmp(optarg, "wave") == 0)
	format = SPRO_SIG_WAVE_FORMAT;
#ifdef SPHERE
      else if (strcasecmp(optarg, "sphere") == 0)
	format = SPRO_SIG_SPHERE_FORMAT;
      else {
	fprintf(stderr, "ssad error -- unknown input signal format %s\n", optarg);
	fprintf(stderr, "              must be one of PCM16, WAVE or SPHERE.\n");
	return(1);
      }
#else
      else {
	fprintf(stderr, "ssad error -- unknown input signal format %s\n", optarg);
	fprintf(stderr, "              must be one of PCM16 or WAVE.\n");
	return(1);
      }
#endif /* SPHERE */
      break;

    case 'f':
      Fs = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --sample-rate (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'x':
      channel = (int)strtol(optarg, &p, 10);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --channel (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'B':
      swap = 1;
      break;

    case 'S':
      ibs = (size_t)strtol(optarg, &p, 10) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --input-bufsize (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
    case 's':
      uselog = 1;
      break;
      
    case 'l':
      fm_l = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --length (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'd':
      fm_d = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --shift (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'w':
      if (strcasecmp(optarg, "hamming") == 0)
	win = SPRO_HAMMING_WINDOW;
      else if (strcasecmp(optarg, "hanning") == 0)
	win = SPRO_HANNING_WINDOW;
      else if (strcasecmp(optarg, "hanning") == 0)
	win = SPRO_BLACKMAN_WINDOW;
      else {
	fprintf(stderr, "ssad error -- unknown window type %s with --window\n", optarg);
	fprintf(stderr, "              must be one of HAMMING, HANNING or BLACKMAN.\n");
	return(1);
      }
      break;

    case 't':
      threshold = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --threshold (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'm':
      minlen = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "ssad error -- invalid argument %s to --min-length (use --help to get usage)\n", optarg);
	return(1);
      }
      break;    

    case 'r':
      if (ofmt == BOTH) 
	fprintf(stderr, ">>>>> WARNING >>>>> overrinding option --reverse with --all in ssad\n");
      else
	ofmt = SPEECH;
      break;

    case 'a':
      if (ofmt == SPEECH)
	fprintf(stderr, ">>>>> WARNING >>>>> overrinding option --reverse with --all in ssad\n");       
      ofmt = BOTH;
      break;

    case 'v':
      trace = 1;
      if (optarg) {
	trace = (int)strtol(optarg, &p, 10);
	if (p == optarg || *p != 0) {
	  fprintf(stderr, "ssad error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      break;

    case 'V':
      version = 1;
      return(0);
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "ssad error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  /* ----- basic argument sanity check ----- */
  if (channel < 1 || channel > 2) {
    fprintf(stderr, "ssad error -- invalid channel number %d (not in [1,2])\n", channel);
    return(1);
  }

  if (fm_l <= 0.0) {
    fprintf(stderr, "ssad error -- invalid frame length %.2f (must be greater than zero)\n", fm_l);
    return(1);
  }

  if (fm_d <= 0.0) {
    fprintf(stderr, "ssad error -- invalid frame shift %.2f (must be greater than zero)\n", fm_d);
    return(1);
  }

  if (fm_l < fm_d)
    fprintf(stderr, ">>>>> WARNING >>>>>> non overlapping frames in ssad\n");

  if (minlen < 0.0) {
    fprintf(stderr, "ssad error -- invalid minimum segment length %.2f (must be greater or equal zero)\n", minlen);
    return(1);
  }

  return(0);
}


#undef _ssad_c_
